# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import sys
from dataclasses import dataclass

import kernel_explorer as ke
import numpy as np
from utils import dtype_to_bytes


def dtype_to_funcs(dtype):
    type_map = {
        "float16": list(filter(lambda x: "MatrixFloatBnb4_half" in x, dir(ke))),
        "float32": list(filter(lambda x: "MatrixFloatBnb4_float" in x, dir(ke))),
    }
    return type_map[dtype]


def dtype_to_funcs_cublas(dtype):
    type_map = {
        "float16": list(filter(lambda x: "GemmBenchmark_half" in x, dir(ke))),
        "float32": list(filter(lambda x: "GemmBenchmark_float" in x, dir(ke))),
    }
    return type_map[dtype]


quant_enums = {"FP4": 0, "NF4": 1}


dtypes = ["float16", "float32"]
quant_types = ["FP4", "NF4"]


@dataclass
class MatrixMulMetric(ke.BandwidthMetric):
    m: int
    n: int
    k: int

    def report(self):
        return (
            f"{self.duration:6.2f} us {self.gbps:5.2f} GB/s {self.dtype} m={self.m} n={self.n} k={self.k} {self.name}"
        )


@dataclass
class MatrixFpBnb4Metric(MatrixMulMetric):
    quant_type: str

    def report(self):
        return (
            f"{self.duration:6.2f} us {self.gbps:5.2f} GB/s"
            f" {self.quant_type} {self.dtype} m={self.m} n={self.n} k={self.k} {self.name}"
        )


def profile_matmul_fp_bnb4_func(qt, m, n, k, dtype, func):
    np.random.seed(0)
    block_size = 64
    numel = n * k
    output = np.random.rand(m, n).astype(dtype)
    a = np.random.rand(m, k).astype(dtype)
    b = np.random.randint(low=0, high=255, size=(numel + 1) // 2).astype("uint8")
    absmax = np.random.rand((numel + block_size - 1) // block_size).astype(dtype)
    quant_map_buffer = np.zeros(16).astype(dtype)

    output_d = ke.DeviceArray(output)
    a_d = ke.DeviceArray(a)
    b_d = ke.DeviceArray(b)
    absmax_d = ke.DeviceArray(absmax)
    quant_map_buffer_d = ke.DeviceArray(quant_map_buffer)
    f = getattr(ke, func)

    my_op = f(output_d, a_d, b_d, absmax_d, quant_map_buffer_d, quant_enums[qt], m, n, k)
    duration_ms = my_op.Profile()
    total_bytes = (m * k + n * k + m * n) * (dtype_to_bytes(dtype))

    ke.report(MatrixFpBnb4Metric(func, dtype, duration_ms, total_bytes, m, n, k, qt))


def profile_gemm_func(m, n, k, dtype, func):
    np.random.seed(0)
    output = np.random.rand(m, n).astype(dtype)
    a = np.random.rand(m, k).astype(dtype)
    b = np.random.rand(k, n).astype(dtype)

    output_d = ke.DeviceArray(output)
    a_d = ke.DeviceArray(a)
    b_d = ke.DeviceArray(b)
    f = getattr(ke, func)
    my_op = f(output_d, a_d, b_d, m, n, k)
    duration_ms = my_op.Profile()
    total_bytes = (m * k + n * k + m * n) * (dtype_to_bytes(dtype))

    ke.report(MatrixMulMetric(func, dtype, duration_ms, total_bytes, m, n, k))


def profile_with_args(qt, m, n, k, dtype, sort):
    with ke.benchmark(sort):
        for func in dtype_to_funcs(dtype):
            profile_matmul_fp_bnb4_func(qt, m, n, k, dtype, func)

        for func in dtype_to_funcs_cublas(dtype):
            profile_gemm_func(m, n, k, dtype, func)


def profile():
    dims_m = [1]
    for qt in quant_types:
        for dt in dtypes:
            for m in dims_m:
                for n, k in ((4096, 4096), (4096, 12288), (12288, 4096)):
                    profile_with_args(qt, m, n, k, dt, False)
                    print()


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    group = parser.add_argument_group("profile with args")
    group.add_argument("m", type=int)
    group.add_argument("n", type=int)
    group.add_argument("k", type=int)
    group.add_argument("quant_type", choices=quant_types)
    group.add_argument("dtype", choices=dtypes)
    group.add_argument("--sort", action="store_true")

    if len(sys.argv) == 1:
        profile()
    else:
        args = parser.parse_args()
        profile_with_args(args.quant_type, args.m, args.n, args.k, args.dtype, args.sort)



import onnx
from onnx import TensorProto, helper

# Loop is so the Tranpose output is used in a subgraph
loop_body = helper.make_graph(
    [
        helper.make_node("Add", ["transpose:0", "loop_state_in"], ["loop_state_out"], "Add1"),
    ],
    "Loop_body",
    [
        helper.make_tensor_value_info("iteration_num", TensorProto.INT64, [1]),
        helper.make_tensor_value_info("subgraph_keep_going_in", TensorProto.BOOL, [1]),
        helper.make_tensor_value_info("loop_state_in", TensorProto.FLOAT, [1]),
    ],
    [
        helper.make_tensor_value_info("subgraph_keep_going_in", TensorProto.BOOL, [1]),
        helper.make_tensor_value_info("loop_state_out", TensorProto.FLOAT, [2, 2, 2]),
    ],
    [],
)

# Create the main graph
graph_proto = helper.make_graph(
    [
        # add a Transpose that can be moved past the Slice
        helper.make_node(
            "Transpose",
            inputs=["input:0"],
            outputs=["transpose:0"],
            name="transpose0",
            perm=[1, 0, 2],
        ),
        helper.make_node(
            "Slice",
            inputs=["transpose:0", "start", "end"],
            outputs=["strided_slice:0"],
            name="slice0",
        ),
        helper.make_node(
            "Squeeze",
            inputs=["strided_slice:0", "start"],
            outputs=["out:0"],
            name="squeeze0",
        ),
        helper.make_node(
            "Loop",
            ["max_trip_count", "subgraph_keep_going_in", "state_var_in"],
            ["out:1"],
            "Loop1",
            body=loop_body,
        ),
    ],
    "Main_graph",
    [
        helper.make_tensor_value_info("input:0", TensorProto.FLOAT, [2, 2, 2]),
        helper.make_tensor_value_info("state_var_in", TensorProto.FLOAT, [1]),
    ],
    [
        helper.make_tensor_value_info("out:0", TensorProto.FLOAT, [2, 2]),
        helper.make_tensor_value_info("out:1", TensorProto.FLOAT, [2, 2, 2]),
    ],
    [
        helper.make_tensor("start", TensorProto.INT64, [1], [0]),
        helper.make_tensor("end", TensorProto.INT64, [1], [1]),
        helper.make_tensor("max_trip_count", TensorProto.INT64, [1], [1]),
        helper.make_tensor("subgraph_keep_going_in", TensorProto.BOOL, [1], [1]),
    ],
)

model = helper.make_model(graph_proto)
onnx.checker.check_model(model, True)
onnx.save(model, "ort_github_issue_10305.onnx")



from pathlib import Path

import onnx
from onnx import helper
from onnx.onnx_pb import TensorProto

nodes = [
    helper.make_node("Shape", ["X"], ["X_shape"]),
    helper.make_node("Gather", ["X_shape", "gather_indices"], ["gather_out"], axis=0),
    helper.make_node("Slice", ["X_shape", "slice_starts", "slice_ends", "slice_axes", "slice_steps"], ["slice_out"]),
]

graph = helper.make_graph(
    nodes,
    "ShapeThenSliceAndGather",
    [  # input
        helper.make_tensor_value_info("X", TensorProto.FLOAT, ["A", "B", "C", "D", "E"]),
    ],
    [  # output
        helper.make_tensor_value_info("gather_out", TensorProto.INT64, None),
        helper.make_tensor_value_info("slice_out", TensorProto.INT64, None),
    ],
    [  # initializer
        helper.make_tensor("gather_indices", TensorProto.INT64, [2], [1, 3]),
        helper.make_tensor("slice_starts", TensorProto.INT64, [1], [-1]),
        helper.make_tensor("slice_ends", TensorProto.INT64, [1], [-6]),
        helper.make_tensor("slice_axes", TensorProto.INT64, [1], [0]),
        helper.make_tensor("slice_steps", TensorProto.INT64, [1], [-2]),
    ],
)

model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 19)])
onnx.save(model, str(Path(__file__).with_suffix(".onnx")))



from enum import Enum

import onnx
from onnx import OperatorSetIdProto, TensorProto, helper


class Format(Enum):
    Format1 = (1,)
    Format2 = (2,)
    Format3 = 3


def generate_model(model_format, model_name, multi_output_add=False, add_output_in_graph_output=False, with_cast=False):
    nodes = []  # LayerNorm subgraph
    if with_cast:
        nodes.extend(
            [
                helper.make_node("Cast", ["ln_in"], ["c_out"], "cast", to=1),
                helper.make_node("ReduceMean", ["c_out"], ["rd1_out"], "reduce1", axes=[-1], keepdims=1),
                helper.make_node("Sub", ["c_out", "rd1_out"], ["sb1_out"], "sub1"),
                helper.make_node("Sub", ["c_out", "rd1_out"], ["sb2_out"], "sub2"),
            ]
        )
    else:
        nodes.extend(
            [
                helper.make_node("ReduceMean", ["ln_in"], ["rd1_out"], "reduce1", axes=[-1], keepdims=1),
                helper.make_node("Sub", ["ln_in", "rd1_out"], ["sb1_out"], "sub1"),
                helper.make_node("Sub", ["ln_in", "rd1_out"], ["sb2_out"], "sub2"),
            ]
        )
    nodes.extend(
        [  # LayerNorm subgraph
            helper.make_node("Pow", ["sb2_out", "pow_in_2"], ["pow_out"], "pow"),
            helper.make_node("ReduceMean", ["pow_out"], ["rd2_out"], "reduce2", axes=[-1], keepdims=1),
            helper.make_node("Add", ["rd2_out", "const_e12"], ["add1_out"], "add1"),
            helper.make_node("Sqrt", ["add1_out"], ["sqrt_out"], "sqrt"),
            helper.make_node("Div", ["sb1_out", "sqrt_out"], ["div_out"], "div1"),
            helper.make_node("Mul", ["gamma", "div_out"], ["mul_out"], "mul"),
            helper.make_node("Add", ["mul_out", "beta"], ["C"], "add0"),
        ]
    )

    initializers = [  # initializers
        helper.make_tensor("pow_in_2", TensorProto.FLOAT, [], [2]),
        helper.make_tensor("const_e12", TensorProto.FLOAT, [], [1e-12]),
        helper.make_tensor("gamma", TensorProto.FLOAT, [4], [1.0, 2.0, 3.0, 4.0]),
        helper.make_tensor("beta", TensorProto.FLOAT, [4], [0.1, 0.2, 0.3, 0.4]),
    ]

    if model_format is Format.Format1:
        nodes.extend(
            [
                helper.make_node("Add", ["A", "bias"], ["add3_out"], "add3"),
                helper.make_node("Add", ["add3_out", "B"], ["ln_in"], "add2"),
            ]
        )
        initializers.extend(
            [
                helper.make_tensor(
                    "bias", TensorProto.FLOAT16 if with_cast else TensorProto.FLOAT, [4], [0.1, 0.2, 0.3, 0.4]
                ),
            ]
        )
    elif model_format is Format.Format2:
        nodes.extend(
            [
                helper.make_node("Add", ["B", "bias"], ["add3_out"], "add3"),
                helper.make_node("Add", ["A", "add3_out"], ["ln_in"], "add2"),
            ]
        )
        initializers.extend(
            [
                helper.make_tensor(
                    "bias", TensorProto.FLOAT16 if with_cast else TensorProto.FLOAT, [4], [0.1, 0.2, 0.3, 0.4]
                ),
            ]
        )
    elif model_format is Format.Format3:
        nodes.extend(
            [
                helper.make_node("Add", ["A", "B"], ["ln_in"], "add2"),
            ]
        )

    if multi_output_add:
        neg_input = "ln_in" if model_format is Format.Format3 else "add3_out"
        nodes.extend([helper.make_node("Neg", [neg_input], ["neg_out"], "neg")])

    graph = helper.make_graph(
        nodes,
        "SkipLayerNorm_format3",  # name
        [  # inputs
            helper.make_tensor_value_info("A", TensorProto.FLOAT16 if with_cast else TensorProto.FLOAT, [16, 32, 4]),
            helper.make_tensor_value_info("B", TensorProto.FLOAT16 if with_cast else TensorProto.FLOAT, [16, 32, 4]),
        ],
        [  # outputs
            helper.make_tensor_value_info("C", TensorProto.FLOAT, [16, 32, 4]),
        ],
        initializers,
    )

    if add_output_in_graph_output:
        extra_output = "ln_in" if model_format is Format.Format3 else "add3_out"
        graph.output.extend(
            [
                helper.make_tensor_value_info(
                    extra_output, TensorProto.FLOAT16 if with_cast else TensorProto.FLOAT, [16, 32, 4]
                )
            ]
        )

    onnxdomain = OperatorSetIdProto()
    onnxdomain.version = 12
    # The empty string ("") or absence of this field implies the operator set that is defined as part of the ONNX specification.
    onnxdomain.domain = ""
    msdomain = OperatorSetIdProto()
    msdomain.version = 1
    msdomain.domain = "com.microsoft"
    opsets = [onnxdomain, msdomain]

    model = helper.make_model(graph, opset_imports=opsets)
    onnx.save(model, model_name)


def generate_skip_layer_norm(with_cast=False):
    suffix = "_with_cast" if with_cast else ""

    generate_model(Format.Format1, f"skip_layer_norm_format1{suffix}.onnx", with_cast=with_cast)
    generate_model(Format.Format2, f"skip_layer_norm_format2{suffix}.onnx", with_cast=with_cast)
    generate_model(Format.Format3, f"skip_layer_norm_format3{suffix}.onnx", with_cast=with_cast)
    generate_model(
        Format.Format1, f"skip_layer_norm_format1_partial{suffix}.onnx", multi_output_add=True, with_cast=with_cast
    )
    generate_model(
        Format.Format2, f"skip_layer_norm_format2_partial{suffix}.onnx", multi_output_add=True, with_cast=with_cast
    )
    generate_model(
        Format.Format3, f"skip_layer_norm_format3_no_fusion{suffix}.onnx", multi_output_add=True, with_cast=with_cast
    )
    generate_model(
        Format.Format1,
        f"skip_layer_norm_format1_graph_output{suffix}.onnx",
        add_output_in_graph_output=True,
        with_cast=with_cast,
    )
    generate_model(
        Format.Format2,
        f"skip_layer_norm_format2_graph_output{suffix}.onnx",
        add_output_in_graph_output=True,
        with_cast=with_cast,
    )
    generate_model(
        Format.Format3,
        f"skip_layer_norm_format3_graph_output{suffix}.onnx",
        add_output_in_graph_output=True,
        with_cast=with_cast,
    )


generate_skip_layer_norm(with_cast=False)
generate_skip_layer_norm(with_cast=True)



#!/usr/bin/env python3
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
from __future__ import annotations

import argparse
import json
import os
import subprocess
import sys

from run_CIs_for_external_pr import get_pipeline_names
from util.platform_helpers import is_windows


class DefaultArgsRawHelpFormatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter):
    pass


def _parse_args():
    parser = argparse.ArgumentParser(
        os.path.basename(__file__),
        formatter_class=DefaultArgsRawHelpFormatter,
        description="""Run the CIs used to validate PRs for the specified branch.

        If not specified, the branch will be inferred (if possible) by running `git branch --show-current`.

        If specified, the `--include` filter is applied first, followed by any `--exclude` filter.
        `--include` and `--exclude` can be specified multiple times to accumulate values to include/exclude.

        Requires the Azure CLI with DevOps extension to be installed.
          Azure CLI: https://learn.microsoft.com/en-us/cli/azure/install-azure-cli
          DevOps extension: https://github.com/Azure/azure-devops-cli-extension

        Configuration:
          Login:`az login`
          Configure ORT repo as default:
            `az devops configure --defaults organization=https://dev.azure.com/onnxruntime project=onnxruntime`

        Example usage:
          List all CIs
            `python run_CIs_for_branch.py --dry-run my/BranchName`
          Run all CIs
            `python run_CIs_for_branch.py my/BranchName`
          Run only Linux CIs
            `python run_CIs_for_branch.py --include linux my/BranchName`
          Exclude training CIs
            `python run_CIs_for_branch.py --exclude training my/BranchName`
          Run non-training Linux CIs
            `python run_CIs_for_branch.py --include linux --exclude training my/BranchName`
        """,
    )

    current_branch = None
    get_branch_result = subprocess.run(["git", "branch", "--show-current"], capture_output=True, text=True, check=False)
    if get_branch_result.returncode == 0:
        current_branch = get_branch_result.stdout.strip()

    parser.add_argument(
        "-i", "--include", action="append", type=str, help="Include CIs that match this string. Case insensitive."
    )
    parser.add_argument(
        "-e", "--exclude", action="append", type=str, help="Exclude CIs that match this string. Case insensitive."
    )
    parser.add_argument("--dry-run", action="store_true", help="Print selected CIs but do not run them.")
    parser.add_argument(
        "branch",
        type=str,
        nargs="?",
        default=current_branch,
        help="Specify the branch to run. Default is current branch if available.",
    )

    args = parser.parse_args()
    if not args.branch:
        raise ValueError("Branch was unable to be inferred and must be specified")

    return args


def _run_az_pipelines_command(command: list[str]):
    try:
        az = "az.cmd" if is_windows() else "az"
        az_output = subprocess.run([az, "pipelines", *command], capture_output=True, text=True, check=True)
    except subprocess.CalledProcessError as cpe:
        print(cpe)
        print(cpe.stderr)
        sys.exit(-1)

    return az_output


def main():
    args = _parse_args()
    branch = args.branch

    # To debug available pipelines:
    # az_out = az_pipelines = _run_az_pipelines_command(["list"])
    # pipeline_info = json.loads(az_out.stdout)
    # print(pipeline_info)

    pipelines = get_pipeline_names()
    pipelines_to_run = []
    if args.include:
        values = [i.lower().strip() for i in args.include]
        for p in pipelines:
            include = False
            for value in values:
                if value in p.lower():
                    include = True
                    break

            if include:
                print(f"Including {p}")
                pipelines_to_run.append(p)
    else:
        pipelines_to_run = pipelines

    if args.exclude:
        values = [e.lower().strip() for e in args.exclude]
        cur_pipelines = pipelines_to_run
        pipelines_to_run = []
        for p in cur_pipelines:
            exclude = False
            for value in values:
                if value in p.lower():
                    exclude = True
                    break

            if exclude:
                print(f"Excluding {p}")
            else:
                pipelines_to_run.append(p)

    print(f"Pipelines to run for {args.branch}:")
    for p in pipelines_to_run:
        print(f"\t{p}")

    if args.dry_run:
        sys.exit(0)

    for pipeline in pipelines_to_run:
        az_out = _run_az_pipelines_command(["run", "--branch", branch, "--name", pipeline])
        run_output = json.loads(az_out.stdout)
        if "id" in run_output:
            build_url = f"https://dev.azure.com/onnxruntime/onnxruntime/_build/results?buildId={run_output['id']}"
            print(f"{pipeline} build results: {build_url}&view=results")
        else:
            raise ValueError("Build id was not found in az output:\n" + run_output)


if __name__ == "__main__":
    main()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""
Train, convert and predict with ONNX Runtime
============================================

This example demonstrates an end to end scenario
starting with the training of a scikit-learn pipeline
which takes as inputs not a regular vector but a
dictionary ``{ int: float }`` as its first step is a
`DictVectorizer <http://scikit-learn.org/stable/modules/generated/sklearn.feature_extraction.DictVectorizer.html>`_.

Train a pipeline
++++++++++++++++

The first step consists in creating a dummy datasets.
"""

import pandas
from sklearn.datasets import make_regression
from sklearn.model_selection import train_test_split

X, y = make_regression(1000, n_targets=1)

X_train, X_test, y_train, y_test = train_test_split(X, y)
X_train_dict = pandas.DataFrame(X_train[:, 1:]).T.to_dict().values()
X_test_dict = pandas.DataFrame(X_test[:, 1:]).T.to_dict().values()

####################################
# We create a pipeline.

from sklearn.ensemble import GradientBoostingRegressor  # noqa: E402
from sklearn.feature_extraction import DictVectorizer  # noqa: E402
from sklearn.pipeline import make_pipeline  # noqa: E402

pipe = make_pipeline(DictVectorizer(sparse=False), GradientBoostingRegressor())

pipe.fit(X_train_dict, y_train)

####################################
# We compute the prediction on the test set
# and we show the confusion matrix.
from sklearn.metrics import r2_score  # noqa: E402

pred = pipe.predict(X_test_dict)
print(r2_score(y_test, pred))

####################################
# Conversion to ONNX format
# +++++++++++++++++++++++++
#
# We use module
# `sklearn-onnx <https://github.com/onnx/sklearn-onnx>`_
# to convert the model into ONNX format.

from skl2onnx import convert_sklearn  # noqa: E402
from skl2onnx.common.data_types import DictionaryType, FloatTensorType, Int64TensorType  # noqa: E402

# initial_type = [('float_input', DictionaryType(Int64TensorType([1]), FloatTensorType([])))]
initial_type = [("float_input", DictionaryType(Int64TensorType([1]), FloatTensorType([])))]
onx = convert_sklearn(pipe, initial_types=initial_type, target_opset=17)
with open("pipeline_vectorize.onnx", "wb") as f:
    f.write(onx.SerializeToString())

##################################
# We load the model with ONNX Runtime and look at
# its input and output.
import onnxruntime as rt  # noqa: E402
from onnxruntime.capi.onnxruntime_pybind11_state import InvalidArgument  # noqa: E402

sess = rt.InferenceSession("pipeline_vectorize.onnx", providers=rt.get_available_providers())

inp, out = sess.get_inputs()[0], sess.get_outputs()[0]
print(f"input name='{inp.name}' and shape={inp.shape} and type={inp.type}")
print(f"output name='{out.name}' and shape={out.shape} and type={out.type}")

##################################
# We compute the predictions.
# We could do that in one call:

try:
    sess.run([out.name], {inp.name: X_test_dict})[0]
except (RuntimeError, InvalidArgument) as e:
    print(e)

#############################
# But it fails because, in case of a DictVectorizer,
# ONNX Runtime expects one observation at a time.
pred_onx = [sess.run([out.name], {inp.name: row})[0][0, 0] for row in X_test_dict]

###############################
# We compare them to the model's ones.
print(r2_score(pred, pred_onx))

#########################
# Very similar. *ONNX Runtime* uses floats instead of doubles,
# that explains the small discrepencies.



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from __future__ import annotations

import os
from typing import TYPE_CHECKING

from onnxruntime.capi import _pybind_state as C

if TYPE_CHECKING:
    from onnxruntime.training.api.module import Module


class Optimizer:
    """Class that provides methods to update the model parameters based on the computed gradients.

    Args:
        optimizer_uri: The path to the optimizer model.
        model: The module to be trained.
    """

    def __init__(self, optimizer_uri: str | os.PathLike, module: Module):
        self._optimizer = C.Optimizer(
            os.fspath(optimizer_uri), module._state._state, module._device, module._session_options
        )

    def step(self) -> None:
        """Updates the model parameters based on the computed gradients.

        This method updates the model parameters by taking a step in the direction of the computed gradients.
        The optimizer used depends on the optimizer model provided.
        """
        self._optimizer.optimizer_step()

    def set_learning_rate(self, learning_rate: float) -> None:
        """Sets the learning rate for the optimizer.

        Args:
            learning_rate: The learning rate to be set.
        """
        self._optimizer.set_learning_rate(learning_rate)

    def get_learning_rate(self) -> float:
        """Gets the current learning rate of the optimizer.

        Returns:
            The current learning rate.
        """
        return self._optimizer.get_learning_rate()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import unittest

import torch
import torch.nn as nn
import torch.nn.functional as F

# Ask lazy backend to use Pytorch's JIT as
# lazy backend's executor.
from torch._lazy.ts_backend import init as init_ts_backend

init_ts_backend()

# Handle ORT dependencies.
import onnxruntime as ort  # noqa: E402
from onnxruntime.capi import _pybind_state as C  # noqa: E402

# Set up ORT as torch.jit's sub-executor.
C.register_ort_as_torch_jit_executor()

# Make computation deterministic.
torch.manual_seed(42)
ort.set_seed(1)


class TestOrtLazyTensor(unittest.TestCase):
    def test_elementwise_model(self):
        def run_elementwise_model():
            # A function to test.
            def elementwise_model(x):
                w = x.relu()
                y = w * w + 1.5
                z = y + x
                p = z * x
                q = p.relu()
                return q

            def run(fun, device, x):
                x = torch.tensor(x, device=device, dtype=torch.float32).requires_grad_()
                y = fun(x)
                y.sum().backward()
                return x, y, x.grad

            # Baseline.
            x, y, g_x = run(elementwise_model, "cpu", [-1.0, 2.0])
            # ORT result.
            x_new, y_new, g_x_new = run(elementwise_model, "lazy", [-1.0, 2.0])

            torch.testing.assert_close(x.to("lazy"), x_new)
            torch.testing.assert_close(y.to("lazy"), y_new)
            torch.testing.assert_close(g_x.to("lazy"), g_x_new)

        for _ in range(5):
            run_elementwise_model()

    def test_mnist_model(self):
        def run_mnist_model():
            class MNISTModel(nn.Module):
                def __init__(self):
                    super().__init__()
                    self.conv1 = nn.Conv2d(1, 32, 3, 1, bias=False)
                    self.conv2 = nn.Conv2d(32, 64, 3, 1, bias=False)
                    self.fc1 = nn.Linear(9216, 128, bias=False)
                    self.fc2 = nn.Linear(128, 10, bias=False)

                def forward(self, x):
                    x = self.conv1(x)
                    x = F.relu(x)
                    x = self.conv2(x)
                    x = F.relu(x)
                    x = F.max_pool2d(x, 2)
                    x = torch.flatten(x, 1)
                    x = self.fc1(x)
                    x = F.relu(x)
                    x = self.fc2(x)
                    output = F.log_softmax(x, dim=1)
                    return output

            def run(model, device, x, y):
                for param in model.parameters():
                    param.grad = None
                model.to(device)
                x = x.to(device)
                y = y.to(device)
                output = model(x)
                loss = F.nll_loss(output, y)
                # return loss
                loss.backward()
                return loss, (param.grad for param in model.parameters())

            x = torch.rand((64, 1, 28, 28), dtype=torch.float32)
            y = torch.randint(0, 9, (64,), dtype=torch.int64)
            model = MNISTModel()

            # Baseline.
            loss, grads = run(model, "cpu", x, y)
            # ORT result.
            loss_new, grads_new = run(model, "lazy", x, y)

            print(f"MNIST loss: {loss} (pytorch), {loss_new} (ort).")
            torch.testing.assert_close(loss.to("lazy"), loss_new, rtol=1e-2, atol=1e-5)
            for g, g_new in zip(grads, grads_new, strict=False):
                torch.testing.assert_close(g.to("lazy"), g_new)

        for _ in range(5):
            run_mnist_model()


if __name__ == "__main__":
    # For a specific model, the first 1 or 2 runs of Pytorch
    # JIT is actual eager mode. As a Pytorch JIT sub-executor,
    # ORT won't be unless we run multiple times. Thus, in each
    # test function, we repeat their core test function multiple times.
    # Here we repeat 5 times because we want to our test similar to
    # training loop.
    # TODO: we should force torch.jit executor to use ORT at the first run.
    unittest.main()



#!/usr/bin/env python
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from __future__ import annotations

import os
import tempfile
import unittest

import numpy as np
import onnx
from op_test_utils import (
    TestDataFeeds,
    check_model_correctness,
    check_op_type_count,
    get_tensor_consumers_and_producers,
)

from onnxruntime.quantization import QuantFormat, QuantType, quantize_static


class TestQDQSlice(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._tmp_model_dir = tempfile.TemporaryDirectory(prefix="ort.qdq.slice_")

        # Note: swap with the commented line if you want to see the models in local test dir.
        cls._tmp_dir_path = cls._tmp_model_dir.name
        # cls._tmp_dir_path = "."

    @classmethod
    def tearDownClass(cls):
        cls._tmp_model_dir.cleanup()

    def build_slice_model(
        self,
        input_shape: list[int],
        input_tensor_type: onnx.TensorProto.DataType,
        starts: list[int],
        ends: list[int],
        axes: list[int] | None = None,
        steps: list[int] | None = None,
    ) -> onnx.ModelProto:
        """
        Returns an onnx.ModelProto with a single Slice operator.
        """
        input_0 = onnx.helper.make_tensor_value_info("input_0", input_tensor_type, input_shape)
        output_0 = onnx.helper.make_tensor_value_info("output_0", input_tensor_type, None)

        initializers = [
            onnx.numpy_helper.from_array(np.array(starts, dtype=np.int64), "starts"),
            onnx.numpy_helper.from_array(np.array(ends, dtype=np.int64), "ends"),
        ]
        slice_input_names = ["input_0", "starts", "ends"]

        if axes:
            initializers.append(onnx.numpy_helper.from_array(np.array(axes, dtype=np.int64), "axes"))
            slice_input_names.append("axes")

        if steps:
            if not axes:
                slice_input_names.append("")  # Empty axes input.
            initializers.append(onnx.numpy_helper.from_array(np.array(steps, dtype=np.int64), "steps"))
            slice_input_names.append("steps")

        slice_node = onnx.helper.make_node("Slice", slice_input_names, ["output_0"], name="Slice0")

        graph = onnx.helper.make_graph(
            [slice_node],
            "SliceGraph",
            [input_0],
            [output_0],
            initializer=initializers,
        )
        opset_imports = [onnx.helper.make_opsetid("", 21)]
        model = onnx.helper.make_model(graph, opset_imports=opset_imports)
        model = onnx.shape_inference.infer_shapes(model)
        onnx.checker.check_model(model, True)
        return model

    def test_qdq_slice_qparams(self):
        """
        Test that QDQ Slice has equal scale/zero-point for its input and output.
        """
        test_configs = [onnx.TensorProto.FLOAT, onnx.TensorProto.FLOAT16]

        for onnx_tensor_type in test_configs:
            with self.subTest(onnx_tensor_type=onnx_tensor_type):
                label = f"{onnx.TensorProto.DataType.Name(onnx_tensor_type)}"
                float_model_path = os.path.join(self._tmp_dir_path, f"slice.{label}.onnx")
                qdq_model_path = os.path.join(self._tmp_dir_path, f"slice.{label}.qdq.onnx")

                input_shape = [2, 4]
                float_model = self.build_slice_model(
                    input_shape=input_shape,
                    input_tensor_type=onnx_tensor_type,
                    starts=[1, 0],
                    ends=[2, 3],
                    axes=None,
                    steps=[1, 2],
                )
                onnx.save_model(float_model, float_model_path)

                # Create a data reader
                np_dtype = onnx.helper.tensor_dtype_to_np_dtype(onnx_tensor_type)
                input_data_list = [
                    {"input_0": np.array([[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0]], dtype=np_dtype)},
                    {"input_0": np.array([[-1.0, -2.0, -3.0, -4.0], [-5.0, -6.0, -7.0, -8.0]], dtype=np_dtype)},
                ]
                data_reader = TestDataFeeds(input_data_list)

                # quantize model to QDQ
                quantize_static(
                    float_model_path,
                    qdq_model_path,
                    data_reader,
                    quant_format=QuantFormat.QDQ,
                    activation_type=QuantType.QUInt8,
                    weight_type=QuantType.QInt8,
                    extra_options={"ForceQuantizeNoInputCheck": True},
                )
                expected_op_counts = {"DequantizeLinear": 2, "QuantizeLinear": 2, "Slice": 1}
                check_op_type_count(self, qdq_model_path, **expected_op_counts)

                data_reader.rewind()
                check_model_correctness(self, float_model_path, qdq_model_path, data_reader.get_next())

                qdq_model = onnx.load_model(qdq_model_path)

                slice_node = next((node for node in qdq_model.graph.node if node.op_type == "Slice"), None)
                self.assertNotEqual(slice_node, None)
                self.assertEqual(slice_node.op_type, "Slice")

                # Get the parent and child nodes of the Slice and check that they are DQ/Q.
                consumers, producers = get_tensor_consumers_and_producers(qdq_model)
                input_dq_node = producers.get(slice_node.input[0], None)
                self.assertNotEqual(input_dq_node, None)
                self.assertEqual(input_dq_node.op_type, "DequantizeLinear")

                output_q_node = consumers.get(slice_node.output[0], [None])[0]
                self.assertNotEqual(output_q_node, None)
                self.assertEqual(output_q_node.op_type, "QuantizeLinear")

                # Check that the Slice's input DQ uses the same scale/zp as the Slice's output Q.
                self.assertEqual(input_dq_node.input[1], output_q_node.input[1])
                self.assertEqual(input_dq_node.input[2], output_q_node.input[2])


if __name__ == "__main__":
    unittest.main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

import argparse
from dataclasses import dataclass

import numpy as np
from benchmark import BenchmarkOp, add_arguments


@dataclass
class OpParam:
    dim1: int
    dim2: int
    dim3: int
    data_type: type


@dataclass
class ModelParam:
    batch_size: int
    seq_len: int
    inter_dim: int
    data_type: type


class BenchmarkFastGelu(BenchmarkOp):
    def __init__(self, args):
        BenchmarkOp.__init__(self, args)

    @classmethod
    def create_inputs_outputs(cls, op_param):
        np.random.seed(0)
        a = np.random.rand(op_param.dim1, op_param.dim2, op_param.dim3).astype(op_param.data_type)
        b = np.random.rand(op_param.dim3).astype(op_param.data_type)
        c = np.random.rand(op_param.dim1, op_param.dim2, op_param.dim3).astype(op_param.data_type)
        inputs = {"A": a, "B": b}
        outputs = {"return_val": c}
        return inputs, outputs

    def create_cases(self):
        model = "models/fast_gelu_fp16.onnx" if self.args.precision == "fp16" else "models/fast_gelu_fp32.onnx"
        data_type = np.float16 if self.args.precision == "fp16" else np.float32
        # bert-large
        model_param = ModelParam(1, 384, 1024 * 4, data_type)
        op_param = OpParam(
            model_param.batch_size,
            model_param.seq_len,
            model_param.inter_dim,
            model_param.data_type,
        )
        self.add_case(op_param, model)

    @classmethod
    def case_profile(cls, op_param, time):
        profile = f"(dim1 dim2 dim3) = ({op_param.dim1} {op_param.dim2} {op_param.dim3}), {time:7.4f} ms"
        return profile


def main():
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    args = parser.parse_args()
    bm = BenchmarkFastGelu(args)
    bm.benchmark()


if __name__ == "__main__":
    main()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: fbs

import flatbuffers
from flatbuffers.compat import import_numpy
np = import_numpy()

# nodes to consider for a runtime optimization
# see corresponding type in onnxruntime/core/graph/runtime_optimization_record.h
class NodesToOptimizeIndices(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = NodesToOptimizeIndices()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsNodesToOptimizeIndices(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)
    @classmethod
    def NodesToOptimizeIndicesBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(buf, offset, b"\x4F\x52\x54\x4D", size_prefixed=size_prefixed)

    # NodesToOptimizeIndices
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # NodesToOptimizeIndices
    def NodeIndices(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # NodesToOptimizeIndices
    def NodeIndicesAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint32Flags, o)
        return 0

    # NodesToOptimizeIndices
    def NodeIndicesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # NodesToOptimizeIndices
    def NodeIndicesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # NodesToOptimizeIndices
    def NumInputs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # NodesToOptimizeIndices
    def NumOutputs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # NodesToOptimizeIndices
    def HasVariadicInput(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # NodesToOptimizeIndices
    def HasVariadicOutput(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # NodesToOptimizeIndices
    def NumVariadicInputs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # NodesToOptimizeIndices
    def NumVariadicOutputs(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

def NodesToOptimizeIndicesStart(builder):
    builder.StartObject(7)

def Start(builder):
    NodesToOptimizeIndicesStart(builder)

def NodesToOptimizeIndicesAddNodeIndices(builder, nodeIndices):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(nodeIndices), 0)

def AddNodeIndices(builder, nodeIndices):
    NodesToOptimizeIndicesAddNodeIndices(builder, nodeIndices)

def NodesToOptimizeIndicesStartNodeIndicesVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)

def StartNodeIndicesVector(builder, numElems: int) -> int:
    return NodesToOptimizeIndicesStartNodeIndicesVector(builder, numElems)

def NodesToOptimizeIndicesAddNumInputs(builder, numInputs):
    builder.PrependUint32Slot(1, numInputs, 0)

def AddNumInputs(builder, numInputs):
    NodesToOptimizeIndicesAddNumInputs(builder, numInputs)

def NodesToOptimizeIndicesAddNumOutputs(builder, numOutputs):
    builder.PrependUint32Slot(2, numOutputs, 0)

def AddNumOutputs(builder, numOutputs):
    NodesToOptimizeIndicesAddNumOutputs(builder, numOutputs)

def NodesToOptimizeIndicesAddHasVariadicInput(builder, hasVariadicInput):
    builder.PrependBoolSlot(3, hasVariadicInput, 0)

def AddHasVariadicInput(builder, hasVariadicInput):
    NodesToOptimizeIndicesAddHasVariadicInput(builder, hasVariadicInput)

def NodesToOptimizeIndicesAddHasVariadicOutput(builder, hasVariadicOutput):
    builder.PrependBoolSlot(4, hasVariadicOutput, 0)

def AddHasVariadicOutput(builder, hasVariadicOutput):
    NodesToOptimizeIndicesAddHasVariadicOutput(builder, hasVariadicOutput)

def NodesToOptimizeIndicesAddNumVariadicInputs(builder, numVariadicInputs):
    builder.PrependUint32Slot(5, numVariadicInputs, 0)

def AddNumVariadicInputs(builder, numVariadicInputs):
    NodesToOptimizeIndicesAddNumVariadicInputs(builder, numVariadicInputs)

def NodesToOptimizeIndicesAddNumVariadicOutputs(builder, numVariadicOutputs):
    builder.PrependUint32Slot(6, numVariadicOutputs, 0)

def AddNumVariadicOutputs(builder, numVariadicOutputs):
    NodesToOptimizeIndicesAddNumVariadicOutputs(builder, numVariadicOutputs)

def NodesToOptimizeIndicesEnd(builder):
    return builder.EndObject()

def End(builder):
    return NodesToOptimizeIndicesEnd(builder)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

from __future__ import annotations

import inspect

import onnx
import torch

from onnxruntime.capi._pybind_state import register_miscellaneous_const_input, register_torch_autograd_function

from ._custom_autograd_function_exporter import register_custom_function_schema_supplementary
from ._utils import get_fully_qualified_class_name

PYTHON_OP_DOMAIN = "com.microsoft"
PYTHON_OP_TYPE = "PythonOp"

PYTHON_OP_ATTRIBUTE_FUNC_NAME = "func_name"
PYTHON_OP_ATTRIBUTE_SAFE_RUN_MODE = "safe_run_mode"
PYTHON_OP_ATTRIBUTE_TRAINING_MODE = "training_mode"


def set_safe_run_mode(model: onnx.ModelProto, allowed_unsafe_run_python_op_names: list[str]) -> onnx.ModelProto:
    # Update safe_run_mode attribute for PythonOp.
    for node in model.graph.node:
        if node.domain == PYTHON_OP_DOMAIN and node.op_type == PYTHON_OP_TYPE:
            func_name = None
            safe_run_mode_attr = None
            for attr in node.attribute:
                if attr.name == PYTHON_OP_ATTRIBUTE_FUNC_NAME:
                    func_name = attr.s.decode("utf-8") if isinstance(attr.s, bytes) else attr.s
                if attr.name == PYTHON_OP_ATTRIBUTE_SAFE_RUN_MODE:
                    safe_run_mode_attr = attr

            if func_name in allowed_unsafe_run_python_op_names:
                if safe_run_mode_attr:
                    node.attribute.remove(safe_run_mode_attr)
                node.attribute.append(onnx.helper.make_attribute(PYTHON_OP_ATTRIBUTE_SAFE_RUN_MODE, 0))

    return model


_PYTHON_OP_INCRE_INDEX = [0]


def make_pythonop_node(
    name_prefix: str,
    inputs: list[
        onnx.ValueInfoProto | int | bool | float | tuple[int, ...] | tuple[bool, ...] | tuple[float, ...] | object
    ],
    outputs: list[onnx.ValueInfoProto],
    func_class: torch.autograd.Function,
    training_mode: int,
    safe_run_mode: int,
) -> onnx.NodeProto:
    assert issubclass(func_class, torch.autograd.Function), "func_class must be a subclass of torch.autograd.Function."

    assert len(inputs) > 0, f"inputs must not be empty for function {func_class}."
    assert len(outputs) > 0, f"outputs must not be empty for function {func_class}."

    all_input_parameters: list[inspect.Parameter] = list(inspect.signature(func_class.forward).parameters.values())

    # Remove the first parameter (ctx) from inspected parameter list.
    assert len(inputs) == len(all_input_parameters) - 1, (
        f"The number of inputs ({len(inputs)}) must match the number of parameters "
        f"({len(all_input_parameters) - 1}) of the forward function."
    )

    func_full_qual_name = get_fully_qualified_class_name(func_class)

    input_tensor_types = []
    input_tensor_ranks = []

    input_bool_scalars = []
    input_bool_scalar_positions = []

    input_int_scalars = []
    input_int_scalar_positions = []

    input_float_scalars = []
    input_float_scalar_positions = []

    input_bool_tuples = []
    input_bool_tuple_positions = []
    input_bool_tuple_begins = []

    input_int_tuples = []
    input_int_tuple_positions = []
    input_int_tuple_begins = []

    input_float_tuples = []
    input_float_tuple_positions = []
    input_float_tuple_begins = []

    input_pointer_scalars = []
    input_pointer_scalar_positions = []

    tensor_args = []
    debug_comment = ""
    cconv = ""
    # Encode inputs to torch.autograd.Function.
    for i, arg in enumerate(inputs):
        if isinstance(arg, onnx.ValueInfoProto):
            # Got a tensor variable.
            tensor_args.append(arg.name)
            input_tensor_types.append(arg.type.tensor_type.elem_type)
            input_tensor_ranks.append(len(arg.type.tensor_type.shape.dim))
            cconv += "d"
            continue

        cconv += "c"

        # Got a non-tensor variable.
        if isinstance(arg, float):
            # A float.
            input_float_scalar_positions.append(i)
            input_float_scalars.append(arg)
            continue
        # bool check MUST be before int check since bool is a subclass of int
        elif isinstance(arg, bool):
            # A bool.
            input_bool_scalar_positions.append(i)
            input_bool_scalars.append(int(arg))
            continue
        elif isinstance(arg, int):
            # A int.
            input_int_scalar_positions.append(i)
            input_int_scalars.append(arg)
            continue

        is_bool_tuple = False
        is_int_tuple = False
        is_float_tuple = False
        if isinstance(arg, tuple) and len(arg) > 0:
            # bool check MUST be before int check since bool is a subclass of int.
            is_bool_tuple = all(isinstance(ele, bool) for ele in arg)
            is_int_tuple = not is_bool_tuple and all(isinstance(ele, int) for ele in arg)
            is_float_tuple = not is_bool_tuple and not is_int_tuple and all(isinstance(ele, float) for ele in arg)

        # Only support tuple of bool, int or float, for other types, handle it as a pointer.
        if is_bool_tuple:
            # A tuple of bool.
            input_bool_tuple_positions.append(i)
            input_bool_tuple_begins.append(len(input_bool_tuples))
            input_bool_tuples.extend([int(ele) for ele in arg])
            continue
        elif is_int_tuple:
            # A tuple of ints.
            input_int_tuple_positions.append(i)
            input_int_tuple_begins.append(len(input_int_tuples))
            input_int_tuples.extend(list(arg))
            continue
        elif is_float_tuple:
            # A tuple of floats.
            input_float_tuple_positions.append(i)
            input_float_tuple_begins.append(len(input_float_tuples))
            input_float_tuples.extend(list(arg))
            continue

        from onnxruntime.training.utils.hooks._statistics_subscriber import _InspectActivation

        is_inspect_activation = func_full_qual_name == get_fully_qualified_class_name(_InspectActivation)
        if is_inspect_activation and isinstance(arg, str):
            # _InspectActivation is a special case where the first argument is a string
            # that is used to determine the activation name to be inspected.
            debug_comment += arg

        # All other inputs are accessed via "pointers".
        input_pointer_scalar_positions.append(i)
        input_pointer_scalars.append(id(arg))

        # For pointer (for example, ProcessGroup passed to PythonOp) needed for PythonOp execution,
        # we append it into a global store to hold a reference (in case it is released after module exported).
        register_miscellaneous_const_input(arg)

    output_tensor_types = []
    output_tensor_ranks = []
    for arg in outputs:
        output_tensor_types.append(arg.type.tensor_type.elem_type)
        output_tensor_ranks.append(len(arg.type.tensor_type.shape.dim))

    attrs = {
        "func_name": func_full_qual_name,
        "input_convention": cconv,
        "input_tensor_types": input_tensor_types,
        "input_tensor_ranks": input_tensor_ranks,
        "output_tensor_types": output_tensor_types,
        "output_tensor_ranks": output_tensor_ranks,
        "training_mode": training_mode,
        "safe_run_mode": safe_run_mode,
        "comment": debug_comment,
    }

    if len(input_bool_scalars) > 0:
        attrs["input_bool_scalars"] = input_bool_scalars
        attrs["input_bool_scalar_positions"] = input_bool_scalar_positions
    if len(input_int_scalars) > 0:
        attrs["input_int_scalars"] = input_int_scalars
        attrs["input_int_scalar_positions"] = input_int_scalar_positions
    if len(input_float_scalars) > 0:
        attrs["input_float_scalars"] = input_float_scalars
        attrs["input_float_scalar_positions"] = input_float_scalar_positions
    if len(input_bool_tuples) > 0:
        attrs["input_bool_tuples"] = input_bool_tuples
        attrs["input_bool_tuple_positions"] = input_bool_tuple_positions
        attrs["input_bool_tuple_begins"] = input_bool_tuple_begins
    if len(input_int_tuples) > 0:
        attrs["input_int_tuples"] = input_int_tuples
        attrs["input_int_tuple_positions"] = input_int_tuple_positions
        attrs["input_int_tuple_begins"] = input_int_tuple_begins
    if len(input_float_tuples) > 0:
        attrs["input_float_tuples"] = input_float_tuples
        attrs["input_float_tuple_positions"] = input_float_tuple_positions
        attrs["input_float_tuple_begins"] = input_float_tuple_begins
    if len(input_pointer_scalars) > 0:
        attrs["input_pointer_scalars"] = input_pointer_scalars
        attrs["input_pointer_scalar_positions"] = input_pointer_scalar_positions

    # Register function with class names.
    register_torch_autograd_function(func_full_qual_name, func_class)

    register_custom_function_schema_supplementary(func_class)

    _PYTHON_OP_INCRE_INDEX[0] += 1
    node_name = f"{name_prefix}_{_PYTHON_OP_INCRE_INDEX[0]}"

    node = onnx.helper.make_node(
        PYTHON_OP_TYPE,
        tensor_args,
        [f"{node_name}_ctx", *[output.name for output in outputs]],
        node_name,  # node name
        "",
        PYTHON_OP_DOMAIN,
        **attrs,
    )

    return node



import itertools
import json
import os
import typing
import unittest
import warnings

import numpy as np
import onnx.helper as oh
from onnx import TensorProto, load
from onnx.numpy_helper import from_array
from onnx.reference import ReferenceEvaluator

import onnxruntime


def has_cuda():
    available_providers = list(onnxruntime.get_available_providers())
    return "CUDAExecutionProvider" in available_providers


def ignore_warnings(warns: list[Warning]) -> typing.Callable:
    def wrapper(fct):
        if warns is None:
            raise AssertionError(f"warns cannot be None for '{fct}'.")

        def call_f(self):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", warns)
                return fct(self)

        return call_f

    return wrapper


class TestScatterPerProvider(unittest.TestCase):
    def assert_exists(self, filename: str):
        assert os.path.exists(filename), f"Unable to find {filename!r}."

    def common_scatter(self, opset, providers, dtype, reduction, expected_names):
        from onnxruntime import InferenceSession, SessionOptions

        op_type = "ScatterElements" if "ScatterElements" in expected_names else "ScatterND"
        ndim = 2 if op_type == "ScatterElements" else 3

        assert dtype in (np.float16, np.float32)
        itype = TensorProto.FLOAT if dtype == np.float32 else TensorProto.FLOAT16
        model = oh.make_model(
            oh.make_graph(
                [
                    oh.make_node("CastLike", ["X", "I"], ["data"]),
                    oh.make_node(
                        op_type,
                        inputs=["data", "indices", "updates"],
                        outputs=["sy"],
                        # axis=0,
                        reduction=reduction,
                    ),
                    oh.make_node("Sub", ["sy", "I"], ["Y"]),
                ],
                "name",
                [
                    oh.make_tensor_value_info("X", TensorProto.FLOAT, [None] * ndim),
                    oh.make_tensor_value_info("indices", TensorProto.INT64, [None, None]),
                    oh.make_tensor_value_info("updates", itype, [None] * ndim),
                ],
                [oh.make_tensor_value_info("Y", itype, [None] * ndim)],
                [from_array(np.array([0], dtype=dtype), name="I")],
            ),
            opset_imports=[oh.make_opsetid("", opset)],
            ir_version=8 if opset <= 18 else 9,
        )

        if not os.path.exists("temp_dump"):
            os.mkdir("temp_dump")
        for name in os.listdir("temp_dump"):
            os.remove(os.path.join("temp_dump", name))

        filename = f"temp_dump/{op_type}_{providers[0]}_{itype}.onnx"
        opts = SessionOptions()
        opts.optimized_model_filepath = filename
        sess = InferenceSession(model.SerializeToString(), opts, providers=providers)
        self.assertTrue(sess is not None)
        self.assert_exists(filename)
        onx = load(filename)
        names = [n.op_type for n in onx.graph.node]
        self.assertEqual(expected_names, names)

        sonx = str(onx).replace(" ", "").replace("\n", "|")
        sexp = 'op_type:"Cast"|attribute{|name:"to"|type:INT|i:%d|}' % itype  # noqa: UP031
        sexp2 = 'op_type:"Cast"|attribute{|name:"to"|i:%d|type:INT|}' % itype  # noqa: UP031
        assert sexp in sonx or sexp2 in sonx, f"Unable to find a substring in {sonx!r}"
        if providers == ["CPUExecutionProvider"]:
            return

        if op_type == "ScatterElements":
            data = np.zeros((3, 3), dtype=np.float32)
            data[0, 0] = 1
            indices = np.array([[1, 0, 2], [0, 2, 1]], dtype=np.int64)
            updates = np.array([[1.0, 1.1, 1.2], [2.0, 2.1, 2.2]], dtype=dtype)
        else:
            data = np.array(
                [
                    [[1, 2, 3, 4], [5, 6, 7, 8], [8, 7, 6, 5], [4, 3, 2, 1]],
                    [[1, 2, 3, 4], [5, 6, 7, 8], [8, 7, 6, 5], [4, 3, 2, 1]],
                    [[8, 7, 6, 5], [4, 3, 2, 1], [1, 2, 3, 4], [5, 6, 7, 8]],
                    [[8, 7, 6, 5], [4, 3, 2, 1], [1, 2, 3, 4], [5, 6, 7, 8]],
                ],
                dtype=np.float32,
            )
            indices = np.array([[0], [2]], dtype=np.int64)
            updates = np.array(
                [
                    [[5, 5, 5, 5], [6, 6, 6, 6], [7, 7, 7, 7], [8, 8, 8, 8]],
                    [[1, 1, 1, 1], [2, 2, 2, 2], [3, 3, 3, 3], [4, 4, 4, 4]],
                ],
                dtype=dtype,
            )
        opts = SessionOptions()
        opts.enable_profiling = True
        opts.optimized_model_filepath = filename
        sess = InferenceSession(model.SerializeToString(), opts, providers=providers)
        got = sess.run(None, {"X": data, "indices": indices, "updates": updates})[0]
        self.assertEqual(got.dtype, updates.dtype)
        prof = sess.end_profiling()

        with open(prof, "r") as f:  # noqa: UP015
            content = f.read()
        js = json.loads(content)

        exe_providers = []
        suffixes = ["_kernel_time", "_fence_before", "_fence_after"]
        rows = []
        for row in js:
            if "args" in row and isinstance(row["args"], dict):
                for k, v in row["args"].items():
                    row[f"args_{k}"] = v
                del row["args"]
            name = row["name"]
            for suf in suffixes:
                if name.endswith(suf):
                    changed = name[: -len(suf)]
                    row["op_name"] = changed
                    break
            rows.append(row)
            exe_providers.append((row.get("args_provider", None), row.get("args_op_name", None)))
        short_list = [(a, b) for a, b in exe_providers if a is not None and b is not None]
        self.assertEqual(short_list, [("CUDAExecutionProvider", o) for o in expected_names])

    @unittest.skipIf(not has_cuda(), reason="cuda not available")
    @ignore_warnings(DeprecationWarning)
    def test_scatterels_cuda(self):
        default_value = [
            "Cast",
            "ScatterElements",
            "Sub",
        ]
        expected = {
            (np.float32, "none"): default_value,
            (np.float16, "none"): default_value,
            (np.float32, "add"): default_value,
            (np.float16, "add"): default_value,
            (np.float32, "mul"): default_value,
            (np.float16, "mul"): default_value,
            (np.float32, "min"): default_value,
            (np.float16, "min"): default_value,
            (np.float32, "max"): default_value,
            (np.float16, "max"): default_value,
        }
        for opset, dtype, reduction in itertools.product(
            [16, 18], [np.float32, np.float16], ["none", "add", "mul", "min", "max"]
        ):
            with self.subTest(dtype=dtype, reduction=reduction, opset=opset):
                self.common_scatter(
                    opset,
                    ["CUDAExecutionProvider"],
                    dtype,
                    reduction,
                    expected[dtype, reduction],
                )

    @unittest.skipIf(not has_cuda(), reason="cuda not available")
    @ignore_warnings(DeprecationWarning)
    def test_scatternd_cuda(self):
        default_value = [
            "Cast",
            "ScatterND",
            "Sub",
        ]
        expected = {
            (np.float32, "none"): default_value,
            (np.float16, "none"): default_value,
            (np.float32, "add"): default_value,
            (np.float16, "add"): default_value,
            (np.float32, "mul"): default_value,
            (np.float16, "mul"): default_value,
            (np.float32, "min"): default_value,
            (np.float16, "min"): default_value,
            (np.float32, "max"): default_value,
            (np.float16, "max"): default_value,
        }
        for opset, dtype, reduction in itertools.product(
            [16, 18], [np.float32, np.float16], ["none", "add", "mul", "min", "max"]
        ):
            with self.subTest(dtype=dtype, reduction=reduction, opset=opset):
                self.common_scatter(
                    opset,
                    ["CUDAExecutionProvider"],
                    dtype,
                    reduction,
                    expected[dtype, reduction],
                )

    @ignore_warnings(DeprecationWarning)
    def test_scatterels_cpu(self):
        default_value = [
            "Cast",
            "ScatterElements",
            "Sub",
        ]
        expected = {
            (np.float32, "none"): default_value,
            (np.float16, "none"): default_value,
            (np.float32, "add"): default_value,
            (np.float16, "add"): default_value,
            (np.float32, "mul"): default_value,
            (np.float16, "mul"): default_value,
            (np.float32, "min"): default_value,
            (np.float16, "min"): default_value,
            (np.float32, "max"): default_value,
            (np.float16, "max"): default_value,
        }
        for opset, dtype, reduction in itertools.product([16, 18], [np.float32], ["none", "add", "mul", "min", "max"]):
            with self.subTest(dtype=dtype, reduction=reduction, opset=opset):
                self.common_scatter(
                    opset,
                    ["CPUExecutionProvider"],
                    dtype,
                    reduction,
                    expected[dtype, reduction],
                )

    @ignore_warnings(DeprecationWarning)
    def test_scatternd_cpu(self):
        default_value = [
            "Cast",
            "ScatterND",
            "Sub",
        ]
        expected = {
            (np.float32, "none"): default_value,
            (np.float16, "none"): default_value,
            (np.float32, "add"): default_value,
            (np.float16, "add"): default_value,
            (np.float32, "mul"): default_value,
            (np.float16, "mul"): default_value,
            (np.float32, "min"): default_value,
            (np.float16, "min"): default_value,
            (np.float32, "max"): default_value,
            (np.float16, "max"): default_value,
        }
        for opset, dtype, reduction in itertools.product([16, 18], [np.float32], ["none", "add", "mul", "min", "max"]):
            with self.subTest(dtype=dtype, reduction=reduction, opset=opset):
                self.common_scatter(
                    opset,
                    ["CPUExecutionProvider"],
                    dtype,
                    reduction,
                    expected[dtype, reduction],
                )

    def _scatternd_standalone_cuda(self, reduction, line):
        model = oh.make_model(
            oh.make_graph(
                [
                    oh.make_node(
                        "ScatterND",
                        inputs=["data", "indices", "updates"],
                        outputs=["y"],
                        reduction=reduction,
                    )
                ],
                "nd",
                [
                    oh.make_tensor_value_info("data", TensorProto.FLOAT, [None, None, None]),
                    oh.make_tensor_value_info("indices", TensorProto.INT64, [None, None]),
                    oh.make_tensor_value_info("updates", TensorProto.FLOAT, [None, None, None]),
                ],
                [oh.make_tensor_value_info("y", TensorProto.FLOAT, [None, None, None])],
            ),
            opset_imports=[oh.make_opsetid("", 18)],
            ir_version=9,
        )

        data = np.full((2, 2, 3), 0.1, dtype=np.float32)
        indices = np.array([[line], [1 - line], [line]], dtype=np.int64)
        updates = (2 ** (np.arange(18) + 1).astype(np.float32).reshape((3, 2, 3))).astype(np.float32)

        feeds = {"data": data, "indices": indices, "updates": updates}
        ref = ReferenceEvaluator(model)
        expected = ref.run(None, feeds)[0]

        providers = (
            [
                ["CUDAExecutionProvider"],
                ["CPUExecutionProvider"],
            ]
            if has_cuda()
            else [["CPUExecutionProvider"]]
        )
        for provider in providers:
            sess = onnxruntime.InferenceSession(model.SerializeToString(), providers=provider)
            got = sess.run(None, feeds)[0]
            self.assertEqual(expected.tolist(), got.tolist())

    def test_scatternd_standalone_cuda(self):
        self._scatternd_standalone_cuda("add", 0)
        self._scatternd_standalone_cuda("add", 1)
        self._scatternd_standalone_cuda("mul", 0)
        self._scatternd_standalone_cuda("mul", 1)
        self._scatternd_standalone_cuda("min", 0)
        self._scatternd_standalone_cuda("min", 1)
        self._scatternd_standalone_cuda("max", 0)
        self._scatternd_standalone_cuda("max", 1)


if __name__ == "__main__":
    unittest.main(verbosity=2)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.  See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Below are test results for Gelu or FastGelu FP32 kernels using CUDA:

Formula             Input(BeforeCast)  MaxDiff     MaxDiff(Optimized)
0(gelu_python)      FP32               2.38E-07    4.77E-07
0(gelu_python)      FP16               0           6.10E-05
1(gelu)             FP32               4.77E-07    0
1(gelu)             FP16               6.10E-05    0
2(erf_gelu)         FP32               2.38E-07    9.54E-07
2(erf_gelu)         FP16               1.22E-04    1.95E-03
3(gelu_new)         FP32               2.38E-07    2.38E-07
3(gelu_new)         FP16               0           0
4(gelu_fast)        FP32               0           2.38E-07
4(gelu_fast)        FP16               0           3.05E-05
5(openai_gelu)      FP32               0           2.38E-07
5(openai_gelu)      FP16               0           3.05E-05

For comparison, CPU has MaxDiff=4.77E-07 for each formula.
"""

import math
import os
import unittest

import torch
from parity_utilities import export_onnx, optimize_onnx, parse_arguments, run_parity
from torch import nn


class Gelu(nn.Module):
    def __init__(self, formula=4, fp32_gelu_op=False):
        super().__init__()
        self.formula = formula
        self.fp32_gelu_op = fp32_gelu_op

    def gelu(self, x):
        if self.formula == 0:
            return x * 0.5 * (1.0 + torch.erf(x / math.sqrt(2.0)))
        elif self.formula == 1:
            return nn.functional.gelu(x)
        elif self.formula == 2:
            # erf_gelu in Megatron: x * 0.5 * (torch.erf(x / 1.41421).to(dtype=x.dtype)+torch.ones_like(x).to(dtype=x.dtype))
            return x * 0.5 * (torch.erf(x / 1.41421).to(dtype=x.dtype) + 1.0)
        elif self.formula == 3:
            # gelu_new in huggingface transformers
            return 0.5 * x * (1.0 + torch.tanh(math.sqrt(2.0 / math.pi) * (x + 0.044715 * torch.pow(x, 3.0))))
        elif self.formula == 4:
            # gelu_fast in huggingface transformers with lower precision in a constant (0.7978845608)
            return 0.5 * x * (1.0 + torch.tanh(x * 0.7978845608 * (1.0 + 0.044715 * x * x)))
        else:
            # openai_gelu in Megatron
            return 0.5 * x * (1.0 + torch.tanh(0.7978845608028654 * x * (1.0 + 0.044715 * x * x)))

    @staticmethod
    def get_fused_op(formula):
        return "Gelu" if formula in [0, 1, 2] else "FastGelu"

    def forward(self, x):
        if self.fp32_gelu_op and x.dtype == torch.float16:
            # This test only evaluates FP32 kernels so add data type cast for input and output.
            casted_output = self.gelu(x.to(torch.float32)).to(torch.float16)
            return (casted_output,)
        else:
            output = self.gelu(x)
            return (output,)


def get_output_names():
    outputs = ["output"]
    return outputs


def run(
    batch_size,
    float16,
    optimized,
    hidden_size,
    device,
    test_cases,
    formula=0,
    sequence_length=2,
    fp32_gelu_op=True,
    verbose=False,
):
    test_name = f"device={device}, float16={float16}, optimized={optimized}, batch_size={batch_size}, sequence_length={sequence_length}, hidden_size={hidden_size}, formula={formula}, fp32_gelu_op={fp32_gelu_op}"
    print(f"\nTesting: {test_name}")

    model = Gelu(formula=formula, fp32_gelu_op=fp32_gelu_op)
    model.eval()
    model.to(device)
    if float16:
        model.half()

    # Do not re-use onnx file from previous test since weights of model are random.
    onnx_model_path = "./temp/gelu_{}_{}.onnx".format(formula, "fp16" if float16 else "fp32")
    export_onnx(model, onnx_model_path, float16, hidden_size, device)

    if optimized:
        optimized_onnx_path = "./temp/gelu_{}_opt_{}.onnx".format(formula, "fp16" if float16 else "fp32")
        use_gpu = float16 and not fp32_gelu_op
        optimize_onnx(
            onnx_model_path,
            optimized_onnx_path,
            Gelu.get_fused_op(formula),
            use_gpu=use_gpu,
            opt_level=2 if use_gpu else None,
            verbose=verbose,
        )
        onnx_path = optimized_onnx_path
    else:
        onnx_path = onnx_model_path

    num_failure = run_parity(
        model,
        onnx_path,
        batch_size,
        hidden_size,
        sequence_length,
        float16,
        device,
        optimized,
        test_cases,
        verbose,
    )

    # clean up onnx file
    os.remove(onnx_model_path)
    if optimized:
        os.remove(onnx_path)

    return num_failure, test_name


class TestGeluParity(unittest.TestCase):
    verbose = False
    optimized = True

    def setUp(self):
        self.test_cases = 100  # Number of test cases per test run
        self.sequence_length = 2
        self.hidden_size = 768
        self.formula_to_test = [0, 1, 2, 3, 4, 5]
        self.formula_must_pass = [
            0,
            1,
            3,
            4,
            5,
        ]  # formula 2 cannot pass precision test.

    def run_test(
        self,
        batch_size,
        float16,
        optimized,
        hidden_size,
        device,
        formula,
        enable_assert=True,
        fp32_gelu_op=True,
        verbose=False,
    ):
        if float16 and device.type == "cpu":  # CPU does not support FP16
            return
        num_failure, test_name = run(
            batch_size,
            float16,
            optimized,
            hidden_size,
            device,
            self.test_cases,
            formula,
            self.sequence_length,
            fp32_gelu_op,
            verbose,
        )
        if enable_assert:
            self.assertTrue(num_failure == 0, "Failed: " + test_name)

    def run_one(self, optimized, device, hidden_size=768, formula=0, verbose=False):
        for batch_size in [4]:
            self.run_test(
                batch_size,
                float16=False,
                optimized=optimized,
                hidden_size=hidden_size,
                device=device,
                formula=formula,
                enable_assert=formula in self.formula_must_pass,
                verbose=verbose,
            )

            self.run_test(
                batch_size,
                float16=True,
                optimized=optimized,
                hidden_size=hidden_size,
                device=device,
                formula=formula,
                enable_assert=formula in self.formula_must_pass,
                fp32_gelu_op=True,
                verbose=verbose,
            )

            self.run_test(
                batch_size,
                float16=True,
                optimized=optimized,
                hidden_size=hidden_size,
                device=device,
                formula=formula,
                enable_assert=formula in self.formula_must_pass,
                fp32_gelu_op=False,
                verbose=verbose,
            )

    def test_cpu(self):
        cpu = torch.device("cpu")
        for i in self.formula_to_test:
            self.run_one(self.optimized, cpu, hidden_size=self.hidden_size, formula=i, verbose=self.verbose)

    def test_cuda(self):
        if not torch.cuda.is_available():
            self.skipTest("test requires GPU and torch+cuda")
        else:
            gpu = torch.device("cuda")
            for i in self.formula_to_test:
                self.run_one(self.optimized, gpu, hidden_size=self.hidden_size, formula=i, verbose=self.verbose)


if __name__ == "__main__":
    args, remaining_args = parse_arguments(namespace_filter=unittest)

    TestGeluParity.verbose = args.log_verbose
    TestGeluParity.optimized = args.optimize

    unittest.main(argv=remaining_args)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

"""
Decompose a complicated op into a series of simple ops.
"simple ops" can be executed in one pass
"""

import sympy
from onnx import GraphProto, NodeProto, TensorProto, helper

from ._utils import get_attribute, get_reduce_info


def _is_half_dtype(dtype: int):
    return dtype in [TensorProto.FLOAT16, TensorProto.BFLOAT16]


class DecomposeDispatch:
    """
    A node does only responsible for a single computation or a type of triton ops.
    For those compound Onnx nodes, like softmax/layernorm/groupnorm, etc., we need to decompose them into a series of
    simple ops.
    """

    def __init__(self):
        self.count = 0

    def __call__(self, node: NodeProto, graph: GraphProto, **kwargs) -> list[NodeProto]:
        op_type = node.op_type
        if not hasattr(self, op_type):
            raise NotImplementedError(f"Not implemented for op type: {op_type}")
        return getattr(self, op_type)(node, graph, **kwargs)

    def __contains__(self, node: NodeProto) -> bool:
        return hasattr(self, node.op_type)

    def _get_unique_var_name(self, prefix):
        self.count += 1
        return prefix + str(self.count)

    def _new_node(self, node_name, op_type, inputs, outputs=None, **kwargs):
        name = self._get_unique_var_name(f"{node_name}_{op_type}_")
        if outputs is None:
            outputs = [f"{name}_out"]
        for idx, output in enumerate(outputs):
            if output is None:
                outputs[idx] = f"{name}_out{idx}"
        return helper.make_node(op_type, inputs, outputs, name, **kwargs), *outputs

    def _get_dtype_and_shape(self, arg_name: str, **kwargs):
        node_arg_infos = kwargs["node_arg_infos"]
        arg_info = node_arg_infos[arg_name]
        return arg_info.dtype, arg_info.shape

    def _decompose_elementwise_precision(self, node: NodeProto, **kwargs):
        x = node.input[0]
        dtype, _ = self._get_dtype_and_shape(x, **kwargs)
        if not _is_half_dtype(dtype):
            return [node]
        node_name = node.name
        y = node.output[0]
        op_type = node.op_type
        inputs = list(node.input)
        cast_nodes = []
        for idx, input in enumerate(inputs):
            dtype, _ = self._get_dtype_and_shape(input, **kwargs)
            if _is_half_dtype(dtype):
                cast_node, cast_out = self._new_node(node_name, "Cast", [input], to=TensorProto.FLOAT)
                cast_nodes.append(cast_node)
                inputs[idx] = cast_out
        op_node, op_out = self._new_node(node_name, op_type, inputs)
        cast_node1, _ = self._new_node(node_name, "Cast", [op_out], outputs=[y], to=dtype)
        return [*cast_nodes, op_node, cast_node1]

    def Exp(self, node: NodeProto, graph: GraphProto, **kwargs):  # noqa: N802
        # pylint: disable=unused-argument
        return self._decompose_elementwise_precision(node, **kwargs)

    def Pow(self, node: NodeProto, graph: GraphProto, **kwargs):  # noqa: N802
        # pylint: disable=unused-argument
        return self._decompose_elementwise_precision(node, **kwargs)

    def Sqrt(self, node: NodeProto, graph: GraphProto, **kwargs):  # noqa: N802
        # pylint: disable=unused-argument
        return self._decompose_elementwise_precision(node, **kwargs)

    def LayerNormalization(self, node: NodeProto, graph: GraphProto, **kwargs):  # noqa: N802
        # pylint: disable=unused-argument
        node_name = node.name
        x = node.input[0]
        w = node.input[1]
        b = node.input[2]
        y = node.output[0]
        mean = node.output[1] if len(node.output) > 1 and node.output[1] else None
        inv_std_dev = node.output[2] if len(node.output) > 2 and node.output[2] else None
        axis = get_attribute(node, "axis", -1)
        epsilon = get_attribute(node, "epsilon", 1e-05)
        xdtype, shape = self._get_dtype_and_shape(x, **kwargs)
        wdtype, _ = self._get_dtype_and_shape(w, **kwargs)
        is_x_half = _is_half_dtype(xdtype)
        is_w_half = _is_half_dtype(wdtype)
        if is_x_half or is_w_half:
            decomposed_nodes = []
            if is_x_half:
                cast_node, x = self._new_node(node_name, "Cast", [x], to=TensorProto.FLOAT)
                decomposed_nodes.append(cast_node)
            if is_w_half:
                cast_node1, w = self._new_node(node_name, "Cast", [w], to=TensorProto.FLOAT)
                cast_node2, b = self._new_node(node_name, "Cast", [b], to=TensorProto.FLOAT)
                decomposed_nodes.append(cast_node1)
                decomposed_nodes.append(cast_node2)
            outputs = [None] if is_w_half else [y]
            if mean is not None:
                outputs.append(mean)
            if inv_std_dev is not None:
                outputs.append(inv_std_dev)
            layer_norm_node_outputs = self._new_node(
                node_name, "LayerNormalization", [x, w, b], outputs=outputs, axis=axis, epsilon=epsilon
            )
            decomposed_nodes.append(layer_norm_node_outputs[0])
            if is_w_half:
                cast_node3, _ = self._new_node(node_name, "Cast", [layer_norm_node_outputs[1]], outputs=[y], to=wdtype)
                decomposed_nodes.append(cast_node3)
            return decomposed_nodes
        rank = len(shape)
        if axis < 0:
            axis += rank
        axes = list(range(axis, rank))
        epsilon_tensor = helper.make_tensor(name="epsilon_const", data_type=xdtype, dims=(1,), vals=[epsilon])
        const_node, const_out = self._new_node(node_name, "Constant", [], value=epsilon_tensor)
        reducemean_node, reducemean_out = self._new_node(node_name, "ReduceMean", [x], outputs=[mean], axes=axes)
        sub_node, sub_out = self._new_node(node_name, "Sub", [x, reducemean_out])
        mul_node, mul_out = self._new_node(node_name, "Mul", [sub_out, sub_out])
        reducemean_node1, reducemean_out1 = self._new_node(node_name, "ReduceMean", [mul_out], axes=axes)
        add_node, add_out = self._new_node(node_name, "Add", [reducemean_out1, const_out])
        rsqrt_node, rsqrt_out = self._new_node(node_name, "Rsqrt", [add_out], outputs=[inv_std_dev])
        mul_node1, mul_out1 = self._new_node(node_name, "Mul", [sub_out, rsqrt_out])
        mul_node2, mul_out2 = self._new_node(node_name, "Mul", [w, mul_out1])
        add_node1, _ = self._new_node(node_name, "Add", [b, mul_out2], outputs=[y])
        return [
            const_node,
            reducemean_node,
            sub_node,
            mul_node,
            reducemean_node1,
            add_node,
            rsqrt_node,
            mul_node1,
            mul_node2,
            add_node1,
        ]

    def LayerNormalizationGrad(self, node: NodeProto, graph: GraphProto, **kwargs):  # noqa: N802
        # pylint: disable=unused-argument
        node_name = node.name
        dy = node.input[0]
        x = node.input[1]
        w = node.input[2]
        mean = node.input[3]
        inv_std_dev = node.input[4]
        dx = node.output[0]
        dw = node.output[1] if len(node.output) > 1 and node.output[1] else None
        db = node.output[2] if len(node.output) > 2 and node.output[2] else None
        axis = get_attribute(node, "axis", -1)
        xdtype, shape = self._get_dtype_and_shape(x, **kwargs)
        wdtype, _ = self._get_dtype_and_shape(w, **kwargs)
        is_x_half = _is_half_dtype(xdtype)
        is_w_half = _is_half_dtype(wdtype)
        if is_x_half or is_w_half:
            decomposed_nodes = []
            if is_x_half:
                cast_node, x = self._new_node(node_name, "Cast", [x], to=TensorProto.FLOAT)
                decomposed_nodes.append(cast_node)
            if is_w_half:
                cast_node1, dy = self._new_node(node_name, "Cast", [dy], to=TensorProto.FLOAT)
                cast_node2, w = self._new_node(node_name, "Cast", [w], to=TensorProto.FLOAT)
                decomposed_nodes.append(cast_node1)
                decomposed_nodes.append(cast_node2)
            outputs = [None] if is_x_half else [dx]
            if dw is not None:
                outputs.append(None if is_w_half else dw)
            if db is not None:
                outputs.append(None if is_w_half else db)
            layer_norm_grad_node_outputs = self._new_node(
                node_name, "LayerNormalizationGrad", [dy, x, w, mean, inv_std_dev], outputs=outputs, axis=axis
            )
            decomposed_nodes.append(layer_norm_grad_node_outputs[0])
            if is_x_half:
                cast_node3, _ = self._new_node(
                    node_name, "Cast", [layer_norm_grad_node_outputs[1]], outputs=[dx], to=xdtype
                )
                decomposed_nodes.append(cast_node3)
            if dw is not None and is_w_half:
                cast_node4, _ = self._new_node(
                    node_name, "Cast", [layer_norm_grad_node_outputs[2]], outputs=[dw], to=wdtype
                )
                decomposed_nodes.append(cast_node4)
            if db is not None and is_w_half:
                cast_node5, _ = self._new_node(
                    node_name, "Cast", [layer_norm_grad_node_outputs[3]], outputs=[db], to=wdtype
                )
                decomposed_nodes.append(cast_node5)
            return decomposed_nodes
        rank = len(shape)
        if axis < 0:
            axis += rank
        axes = list(range(axis, rank))
        sub_node, sub_out = self._new_node(node_name, "Sub", [x, mean])
        mul_node, mul_out = self._new_node(node_name, "Mul", [sub_out, inv_std_dev])
        mul_node1, mul_out1 = self._new_node(node_name, "Mul", [w, dy])
        mul_node2, mul_out2 = self._new_node(node_name, "Mul", [mul_out, mul_out1])
        reducemean_node, reducemean_out = self._new_node(node_name, "ReduceMean", [mul_out2], axes=axes)
        reducemean_node1, reducemean_out1 = self._new_node(node_name, "ReduceMean", [mul_out1], axes=axes)
        mul_node3, mul_out3 = self._new_node(node_name, "Mul", [reducemean_out, mul_out])
        add_node, add_out = self._new_node(node_name, "Add", [mul_out3, reducemean_out1])
        sub_node1, sub_out1 = self._new_node(node_name, "Sub", [mul_out1, add_out])
        mul_node4, _ = self._new_node(node_name, "Mul", [sub_out1, inv_std_dev], outputs=[dx])
        decomposed_nodes = [
            sub_node,
            mul_node,
            mul_node1,
            mul_node2,
            reducemean_node,
            reducemean_node1,
            mul_node3,
            add_node,
            sub_node1,
            mul_node4,
        ]
        dw_axes = list(range(axis))
        if dw is not None:
            mul_node5, mul_out5 = self._new_node(node_name, "Mul", [dy, mul_out])
            reducesum_node, _ = self._new_node(
                node_name, "ReduceSum", [mul_out5], outputs=[dw], axes=dw_axes, keepdims=0
            )
            decomposed_nodes.extend([mul_node5, reducesum_node])
        if db is not None:
            reducesum_node1, _ = self._new_node(node_name, "ReduceSum", [dy], outputs=[db], axes=dw_axes, keepdims=0)
            decomposed_nodes.append(reducesum_node1)
        return decomposed_nodes

    def Softmax(self, node: NodeProto, graph: GraphProto, **kwargs):  # noqa: N802
        # pylint: disable=unused-argument
        node_name = node.name
        x = node.input[0]
        y = node.output[0]
        axis = get_attribute(node, "axis", -1)
        dtype, _ = self._get_dtype_and_shape(x, **kwargs)
        if _is_half_dtype(dtype):
            cast_node, x = self._new_node(node_name, "Cast", [x], to=TensorProto.FLOAT)
            softmax_node, softmax_out = self._new_node(node_name, "Softmax", [x], axis=axis)
            cast_node1, _ = self._new_node(node_name, "Cast", [softmax_out], outputs=[y], to=dtype)
            return [cast_node, softmax_node, cast_node1]
        max_node, max_out = self._new_node(node_name, "ReduceMax", [x], axes=[axis])
        sub_node, sub_out = self._new_node(node_name, "Sub", [x, max_out])
        exp_node, exp_out = self._new_node(node_name, "Exp", [sub_out])
        sum_node, sum_out = self._new_node(node_name, "ReduceSum", [exp_out], axes=[axis])
        div_node, _ = self._new_node(node_name, "Div", [exp_out, sum_out], outputs=[y])
        return [max_node, sub_node, exp_node, sum_node, div_node]

    def SoftmaxGrad_13(self, node: NodeProto, graph: GraphProto, **kwargs):  # noqa: N802
        # pylint: disable=unused-argument
        node_name = node.name
        dy = node.input[0]
        y = node.input[1]
        dx = node.output[0]
        axis = get_attribute(node, "axis", -1)
        dtype, _ = self._get_dtype_and_shape(dy, **kwargs)
        if _is_half_dtype(dtype):
            cast_node, dy = self._new_node(node_name, "Cast", [dy], to=TensorProto.FLOAT)
            cast_node1, y = self._new_node(node_name, "Cast", [y], to=TensorProto.FLOAT)
            softmax_grad_node, softmax_grad_out = self._new_node(node_name, "SoftmaxGrad_13", [dy, y], axis=axis)
            cast_node2, _ = self._new_node(node_name, "Cast", [softmax_grad_out], outputs=[dx], to=dtype)
            return [cast_node, cast_node1, softmax_grad_node, cast_node2]
        mul_node, mul_out = self._new_node(node_name, "Mul", [dy, y])
        sum_node, sum_out = self._new_node(node_name, "ReduceSum", [mul_out], axes=[axis])
        mul_node1, mul_out1 = self._new_node(node_name, "Mul", [y, sum_out])
        sub_node, _ = self._new_node(node_name, "Sub", [mul_out, mul_out1], outputs=[dx])
        return [mul_node, sum_node, mul_node1, sub_node]

    # We support to codegen for reduce Ops that are reducing on contiguous axes. If it's not, we need to decompose
    # it to multiple reduce Ops.
    def _decompose_reduce_axes(self, node: NodeProto, graph: GraphProto, **kwargs):
        node_name = node.name
        op_type = node.op_type
        x = node.input[0]
        y = node.output[0]
        _, shape = self._get_dtype_and_shape(x, **kwargs)
        rank = len(shape)
        keep_dims, axes = get_reduce_info(node, graph, rank)
        if len(axes) == 0:
            identity_node, _ = self._new_node(node_name, "Identity", [x], outputs=[y])
            return [identity_node]
        splited_axes = []
        end = len(axes)
        start = end - 1
        while True:
            while start > 0 and axes[start] == axes[start - 1] + 1:
                start -= 1
            splited_axes.append(axes[start:end])
            if start == 0:
                break
            end = start
            start = end - 1
        if len(splited_axes) == 1:
            if len(node.input) <= 1:
                return [node]
            reduce_node, _ = self._new_node(node_name, op_type, [x], outputs=[y], axes=axes, keepdims=keep_dims)
            return [reduce_node]
        result = []
        for idx, axes in enumerate(splited_axes):
            outputs = [y] if idx == len(splited_axes) - 1 else None
            reduce_node, x = self._new_node(node_name, op_type, [x], outputs=outputs, axes=axes, keepdims=keep_dims)
            result.append(reduce_node)
        return result

    def _decompose_reduce_precision(self, node: NodeProto, graph: GraphProto, **kwargs):
        x = node.input[0]
        dtype, shape = self._get_dtype_and_shape(x, **kwargs)
        if not _is_half_dtype(dtype):
            return [node]
        node_name = node.name
        rank = len(shape)
        keep_dims, axes = get_reduce_info(node, graph, rank)
        y = node.output[0]
        cast_node, x = self._new_node(node_name, "Cast", [x], to=TensorProto.FLOAT)
        reduce_node, reduce_out = self._new_node(node_name, node.op_type, [x], axes=axes, keepdims=keep_dims)
        cast_node1, _ = self._new_node(node_name, "Cast", [reduce_out], outputs=[y], to=dtype)
        return [cast_node, reduce_node, cast_node1]

    def ReduceMax(self, node: NodeProto, graph: GraphProto, **kwargs):  # noqa: N802
        return self._decompose_reduce_axes(node, graph, **kwargs)

    def ReduceMin(self, node: NodeProto, graph: GraphProto, **kwargs):  # noqa: N802
        return self._decompose_reduce_axes(node, graph, **kwargs)

    def ReduceSum(self, node: NodeProto, graph: GraphProto, **kwargs):  # noqa: N802
        precision_decompose_result = self._decompose_reduce_precision(node, graph, **kwargs)
        # The decompose process will be called recursively, if it's already a decomposed result, just return.
        if len(precision_decompose_result) != 1 or precision_decompose_result[0] != node:
            return precision_decompose_result
        return self._decompose_reduce_axes(node, graph, **kwargs)

    def ReduceMean(self, node: NodeProto, graph: GraphProto, **kwargs):  # noqa: N802
        precision_decompose_result = self._decompose_reduce_precision(node, graph, **kwargs)
        # The decompose process will be called recursively, if it's already a decomposed result, just return.
        if len(precision_decompose_result) != 1 or precision_decompose_result[0] != node:
            return precision_decompose_result
        axes_decompose_result = self._decompose_reduce_axes(node, graph, **kwargs)
        # The decompose process will be called recursively, if it's already a decomposed result, just return.
        if len(axes_decompose_result) != 1 or axes_decompose_result[0] != node:
            return axes_decompose_result
        node_name = node.name
        x = node.input[0]
        y = node.output[0]
        dtype, shape = self._get_dtype_and_shape(x, **kwargs)
        rank = len(shape)
        keep_dims, axes = get_reduce_info(node, graph, rank)
        sum_node, sum_out = self._new_node(node_name, "ReduceSum", [x], axes=axes, keepdims=keep_dims)
        # If it's not concrete shape, we need add more Ops such as Shape, Gather to get the dim value,
        # which is not supported yet.
        assert all(shape[axis].is_number for axis in axes)
        denominator = int(sympy.prod([shape[axis] for axis in axes]))
        denominator_tensor = helper.make_tensor(
            name=f"{node_name}_denominator",
            dims=(),
            data_type=dtype,
            vals=[denominator],
        )
        denominator_node, denominator_out = self._new_node(node_name, "Constant", [], value=denominator_tensor)
        div_node, _ = self._new_node(node_name, "Div", [sum_out, denominator_out], outputs=[y])
        return [sum_node, denominator_node, div_node]



class LossScaler:
    r"""Base class for implementing custom loss scaler strategies

    Once the scaler is configured, no user intervention is needed to update loss scale during training.

    Note:
        This class should never be instantiated, but used as an abstract class for custom loss scaling strategy.
    """

    def __init__(self, loss_scale):
        assert isinstance(loss_scale, (int, float)) and loss_scale > 0, "'loss_scale' must be a positive float"
        self._input_name = None
        self._loss_scale = float(loss_scale)
        self._initial_loss_scale = float(loss_scale)

    @property
    def input_name(self):
        return self._input_name

    @input_name.setter
    def input_name(self, input_name):
        assert isinstance(input_name, str), "'input_name' must be a string"
        assert input_name is None or len(input_name) > 0, "'input_name' cannot be empty"
        self._input_name = input_name

    @property
    def loss_scale(self):
        return self._loss_scale

    @loss_scale.setter
    def loss_scale(self, loss_scale):
        assert isinstance(loss_scale, (int, float)) and loss_scale > 0, "'loss_scale' must be a positive float"
        self._loss_scale = float(loss_scale)

    def reset(self):
        r"""Resets loss scaler internal state"""
        self._loss_scale = self._initial_loss_scale

    def update(self, train_step_info):
        r"""Updates loss based on user input and training session info

        Args:
            train_step_info (TrainStepInfo): last step state information

        Returns:
            Updated loss scale (float)
        """
        raise NotImplementedError


class DynamicLossScaler(LossScaler):
    r"""Default implementation for :py:class:`.LossScaler` class used for mixed precision

    This loss scaler works by assuming an initial scale, which is doubled every time a certain number of
    (stable) training steps are performed without exploding gradients (overflow or reach infinity).
    When at least one of the gradients explode, loss scale is divided by 2.

    Users can use this class in two ways:

        1. Enable mixed precision and not setting a loss scaler class. Default values are used
        2. Enable mixed precision and instantiate this class to override default arguments

    Static loss scaling can be achieved by setting :py:attr:`.automatic_update` to :py:obj:`False`
    and not performing manual :py:meth:`update` in train loop.

    Args:
        automatic_update (bool, default is False): boolean switch that allows :py:meth:`ORTTrainer.train_step`
            to automatically perform loss scaling. If False, an explicit call to :py:meth:`.update` must be done by the user,
            otherwise static loss scaling is performed.
        loss_scale (default is 1 << 16): A float that represents current loss scale
        up_scale_window (int, default is 2000): number of stable train steps before doubling loss scale
        min_loss_scale (float, default is 1): min value for the loss scale. Used when loss scale is decreased
        max_loss_scale (float, default is 1 << 24): max value for the loss scale. Used when loss scale is increased

    Example with default values:
        .. code-block:: python

            scaler1 = amp.DynamicLossScaler()
            print(f'Default loss scale is {scaler1.loss_scale}')

    Example with user specified values:
        .. code-block:: python

            scaler2 = amp.DynamicLossScaler(loss_scale=1<<8)
            print(f'Custom loss scale is {scaler2.loss_scale}')
    """

    def __init__(
        self,
        automatic_update=True,
        loss_scale=float(1 << 16),
        up_scale_window=2000,
        min_loss_scale=1.0,
        max_loss_scale=float(1 << 24),
    ):
        super().__init__(loss_scale)
        self.automatic_update = automatic_update
        self.up_scale_window = up_scale_window
        self.min_loss_scale = min_loss_scale
        self.max_loss_scale = max_loss_scale
        self._stable_steps_count = 0

    def reset(self):
        super().reset()
        self._stable_steps_count = 0

    def update(self, train_step_info):
        if not self.automatic_update:
            return self.loss_scale

        if train_step_info.all_finite:
            self._stable_steps_count += 1

            if self._stable_steps_count >= self.up_scale_window:
                self.loss_scale = min(self.max_loss_scale, self.loss_scale * 2)
                self._stable_steps_count = 0
        else:
            self.loss_scale = max(self.min_loss_scale, self.loss_scale / 2)
            self._stable_steps_count = 0
        return self.loss_scale



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# pylint: disable=C0411,C0412,C0413

"""

.. _l-logreg-example-speed:

Train, convert and predict with ONNX Runtime
============================================

This example demonstrates an end to end scenario
starting with the training of a machine learned model
to its use in its converted from.

Train a logistic regression
+++++++++++++++++++++++++++

The first step consists in retrieving the iris datset.
"""

from sklearn.datasets import load_iris
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import train_test_split

iris = load_iris()
X, y = iris.data, iris.target
X_train, X_test, y_train, y_test = train_test_split(X, y)

####################################
# Then we fit a model.

clr = LogisticRegression()
clr.fit(X_train, y_train)

####################################
# We compute the prediction on the test set
# and we show the confusion matrix.
from sklearn.metrics import confusion_matrix  # noqa: E402

pred = clr.predict(X_test)
print(confusion_matrix(y_test, pred))

####################################
# Conversion to ONNX format
# +++++++++++++++++++++++++
#
# We use module
# `sklearn-onnx <https://github.com/onnx/sklearn-onnx>`_
# to convert the model into ONNX format.

from skl2onnx import convert_sklearn  # noqa: E402
from skl2onnx.common.data_types import FloatTensorType  # noqa: E402

initial_type = [("float_input", FloatTensorType([None, 4]))]
onx = convert_sklearn(clr, initial_types=initial_type)
with open("logreg_iris.onnx", "wb") as f:
    f.write(onx.SerializeToString())

##################################
# We load the model with ONNX Runtime and look at
# its input and output.

import onnxruntime as rt  # noqa: E402

sess = rt.InferenceSession("logreg_iris.onnx", providers=rt.get_available_providers())

print(f"input name='{sess.get_inputs()[0].name}' and shape={sess.get_inputs()[0].shape}")
print(f"output name='{sess.get_outputs()[0].name}' and shape={sess.get_outputs()[0].shape}")

##################################
# We compute the predictions.

input_name = sess.get_inputs()[0].name
label_name = sess.get_outputs()[0].name

import numpy  # noqa: E402

pred_onx = sess.run([label_name], {input_name: X_test.astype(numpy.float32)})[0]
print(confusion_matrix(pred, pred_onx))

###################################
# The prediction are perfectly identical.
#
# Probabilities
# +++++++++++++
#
# Probabilities are needed to compute other
# relevant metrics such as the ROC Curve.
# Let's see how to get them first with
# scikit-learn.

prob_sklearn = clr.predict_proba(X_test)
print(prob_sklearn[:3])

#############################
# And then with ONNX Runtime.
# The probabilies appear to be

prob_name = sess.get_outputs()[1].name
prob_rt = sess.run([prob_name], {input_name: X_test.astype(numpy.float32)})[0]

import pprint  # noqa: E402

pprint.pprint(prob_rt[0:3])

###############################
# Let's benchmark.
from timeit import Timer  # noqa: E402


def speed(inst, number=5, repeat=10):
    timer = Timer(inst, globals=globals())
    raw = numpy.array(timer.repeat(repeat, number=number))
    ave = raw.sum() / len(raw) / number
    mi, ma = raw.min() / number, raw.max() / number
    print(f"Average {ave:1.3g} min={mi:1.3g} max={ma:1.3g}")
    return ave


print("Execution time for clr.predict")
speed("clr.predict(X_test)")

print("Execution time for ONNX Runtime")
speed("sess.run([label_name], {input_name: X_test.astype(numpy.float32)})[0]")

###############################
# Let's benchmark a scenario similar to what a webservice
# experiences: the model has to do one prediction at a time
# as opposed to a batch of prediction.


def loop(X_test, fct, n=None):
    nrow = X_test.shape[0]
    if n is None:
        n = nrow
    for i in range(n):
        im = i % nrow
        fct(X_test[im : im + 1])


print("Execution time for clr.predict")
speed("loop(X_test, clr.predict, 50)")


def sess_predict(x):
    return sess.run([label_name], {input_name: x.astype(numpy.float32)})[0]


print("Execution time for sess_predict")
speed("loop(X_test, sess_predict, 50)")

#####################################
# Let's do the same for the probabilities.

print("Execution time for predict_proba")
speed("loop(X_test, clr.predict_proba, 50)")


def sess_predict_proba(x):
    return sess.run([prob_name], {input_name: x.astype(numpy.float32)})[0]


print("Execution time for sess_predict_proba")
speed("loop(X_test, sess_predict_proba, 50)")

#####################################
# This second comparison is better as
# ONNX Runtime, in this experience,
# computes the label and the probabilities
# in every case.

##########################################
# Benchmark with RandomForest
# +++++++++++++++++++++++++++
#
# We first train and save a model in ONNX format.
from sklearn.ensemble import RandomForestClassifier  # noqa: E402

rf = RandomForestClassifier(n_estimators=10)
rf.fit(X_train, y_train)

initial_type = [("float_input", FloatTensorType([1, 4]))]
onx = convert_sklearn(rf, initial_types=initial_type)
with open("rf_iris.onnx", "wb") as f:
    f.write(onx.SerializeToString())

###################################
# We compare.

sess = rt.InferenceSession("rf_iris.onnx", providers=rt.get_available_providers())


def sess_predict_proba_rf(x):
    return sess.run([prob_name], {input_name: x.astype(numpy.float32)})[0]


print("Execution time for predict_proba")
speed("loop(X_test, rf.predict_proba, 50)")

print("Execution time for sess_predict_proba")
speed("loop(X_test, sess_predict_proba_rf, 50)")

##################################
# Let's see with different number of trees.

measures = []

for n_trees in range(5, 51, 5):
    print(n_trees)
    rf = RandomForestClassifier(n_estimators=n_trees)
    rf.fit(X_train, y_train)
    initial_type = [("float_input", FloatTensorType([1, 4]))]
    onx = convert_sklearn(rf, initial_types=initial_type)
    with open(f"rf_iris_{n_trees}.onnx", "wb") as f:
        f.write(onx.SerializeToString())
    sess = rt.InferenceSession(f"rf_iris_{n_trees}.onnx", providers=rt.get_available_providers())

    def sess_predict_proba_loop(x):
        return sess.run([prob_name], {input_name: x.astype(numpy.float32)})[0]  # noqa: B023

    tsk = speed("loop(X_test, rf.predict_proba, 25)", number=5, repeat=4)
    trt = speed("loop(X_test, sess_predict_proba_loop, 25)", number=5, repeat=4)
    measures.append({"n_trees": n_trees, "sklearn": tsk, "rt": trt})

from pandas import DataFrame  # noqa: E402

df = DataFrame(measures)
ax = df.plot(x="n_trees", y="sklearn", label="scikit-learn", c="blue", logy=True)
df.plot(x="n_trees", y="rt", label="onnxruntime", ax=ax, c="green", logy=True)
ax.set_xlabel("Number of trees")
ax.set_ylabel("Prediction time (s)")
ax.set_title("Speed comparison between scikit-learn and ONNX Runtime\nFor a random forest on Iris dataset")
ax.legend()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.  See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import os
import shutil
import unittest

import numpy as np
import pytest
from parity_utilities import find_transformers_source

if find_transformers_source():
    from compare_bert_results import run_test
    from fusion_options import FusionOptions
    from optimizer import optimize_model
else:
    from onnxruntime.transformers.compare_bert_results import run_test
    from onnxruntime.transformers.fusion_options import FusionOptions
    from onnxruntime.transformers.optimizer import optimize_model

if find_transformers_source(["models", "stable_diffusion"]):
    from optimize_pipeline import main as optimize_stable_diffusion
else:
    from onnxruntime.transformers.models.stable_diffusion.optimize_pipeline import main as optimize_stable_diffusion


TINY_MODELS = {
    "stable-diffusion": "hf-internal-testing/tiny-stable-diffusion-torch",
    "stable-diffusion-xl": "echarlaix/tiny-random-stable-diffusion-xl",
    "stable-diffusion-3": "optimum-internal-testing/tiny-random-stable-diffusion-3",
    "flux": "tlwu/tiny-random-flux",
}


class TestStableDiffusionOptimization(unittest.TestCase):
    def verify_node_count(self, onnx_model, expected_node_count, test_name):
        for op_type, count in expected_node_count.items():
            if len(onnx_model.get_nodes_by_op_type(op_type)) != count:
                print(f"Counters is not expected in test: {test_name}")
                for op, counter in expected_node_count.items():
                    print(f"{op}: {len(onnx_model.get_nodes_by_op_type(op))} expected={counter}")

                self.assertEqual(len(onnx_model.get_nodes_by_op_type(op_type)), count)

    def verify_clip_optimizer(self, clip_onnx_path, optimized_clip_onnx_path, expected_counters, float16=False):
        fusion_options = FusionOptions("clip")
        m = optimize_model(
            clip_onnx_path,
            model_type="clip",
            num_heads=0,
            hidden_size=0,
            opt_level=0,
            optimization_options=fusion_options,
            use_gpu=True,
        )
        self.verify_node_count(m, expected_counters, "test_clip")

        if float16:
            m.convert_float_to_float16(
                keep_io_types=True,
            )
        print(m.get_operator_statistics())
        m.save_model_to_file(optimized_clip_onnx_path)

        threshold = 1e-2 if float16 else 3e-3
        max_abs_diff, passed = run_test(
            clip_onnx_path,
            optimized_clip_onnx_path,
            output_dir=None,
            batch_size=1,
            sequence_length=77,
            use_gpu=True,
            test_cases=10,
            seed=1,
            verbose=False,
            rtol=1e-1,
            atol=threshold,
            input_ids_name="input_ids",
            segment_ids_name=None,
            input_mask_name=None,
            mask_type=0,
        )

        self.assertLess(max_abs_diff, threshold)
        self.assertTrue(passed)

    @pytest.mark.slow
    def test_clip_sd(self):
        save_directory = "tiny-random-stable-diffusion"
        if os.path.exists(save_directory):
            shutil.rmtree(save_directory, ignore_errors=True)

        model_type = "stable-diffusion"
        model_name = TINY_MODELS[model_type]

        from optimum.onnxruntime import ORTStableDiffusionPipeline

        base = ORTStableDiffusionPipeline.from_pretrained(model_name, export=True)
        base.save_pretrained(save_directory)

        clip_onnx_path = os.path.join(save_directory, "text_encoder", "model.onnx")
        optimized_clip_onnx_path = os.path.join(save_directory, "text_encoder", "opt.onnx")
        self.verify_clip_optimizer(
            clip_onnx_path,
            optimized_clip_onnx_path,
            expected_counters={
                "EmbedLayerNormalization": 0,
                "Attention": 5,
                "SkipLayerNormalization": 10,
                "LayerNormalization": 1,
                "Gelu": 0,
                "BiasGelu": 0,
            },
            float16=True,
        )


class TestStableDiffusionOrFluxPipelineOptimization(unittest.TestCase):
    def verify_pipeline_optimization(
        self,
        model_name,
        export_onnx_dir,
        optimized_onnx_dir,
        expected_op_counters,
        is_float16,
        atol,
        disable_group_norm=False,
    ):
        from optimum.onnxruntime import ORTPipelineForText2Image

        if os.path.exists(export_onnx_dir):
            shutil.rmtree(export_onnx_dir, ignore_errors=True)

        baseline = ORTPipelineForText2Image.from_pretrained(model_name, export=True, provider="CUDAExecutionProvider")
        if not os.path.exists(export_onnx_dir):
            baseline.save_pretrained(export_onnx_dir)

        argv = [
            "--input",
            export_onnx_dir,
            "--output",
            optimized_onnx_dir,
            "--overwrite",
            "--disable_bias_splitgelu",
        ]

        if disable_group_norm:
            argv.append("--disable_group_norm")

        if is_float16:
            argv.append("--float16")

        op_counters = optimize_stable_diffusion(argv)
        print(op_counters)

        for name in expected_op_counters:
            self.assertIn(name, op_counters)
            for op, count in expected_op_counters[name].items():
                self.assertIn(op, op_counters[name])
                self.assertEqual(op_counters[name][op], count, f"Expected {count} {op} in {name}")

        treatment = ORTPipelineForText2Image.from_pretrained(optimized_onnx_dir, provider="CUDAExecutionProvider")
        batch_size, num_images_per_prompt, height, width = 1, 1, 64, 64
        inputs = {
            "prompt": ["starry night by van gogh"] * batch_size,
            "num_inference_steps": 20,
            "num_images_per_prompt": num_images_per_prompt,
            "height": height,
            "width": width,
            "output_type": "np",
        }

        seed = 123
        np.random.seed(seed)
        import torch

        baseline_outputs = baseline(**inputs, generator=torch.Generator(device="cuda").manual_seed(seed))

        np.random.seed(seed)
        treatment_outputs = treatment(**inputs, generator=torch.Generator(device="cuda").manual_seed(seed))

        self.assertTrue(np.allclose(baseline_outputs.images[0], treatment_outputs.images[0], atol=atol))

    @pytest.mark.slow
    def test_sd(self):
        """This tests optimization of stable diffusion 1.x pipeline"""
        model_name = TINY_MODELS["stable-diffusion"]

        expected_op_counters = {
            "unet": {
                "Attention": 6,
                "MultiHeadAttention": 6,
                "LayerNormalization": 6,
                "SkipLayerNormalization": 12,
                "BiasSplitGelu": 0,
                "GroupNorm": 0,
                "SkipGroupNorm": 0,
                "NhwcConv": 47,
                "BiasAdd": 0,
            },
            "vae_encoder": {"Attention": 0, "GroupNorm": 0, "SkipGroupNorm": 0, "NhwcConv": 13},
            "vae_decoder": {"Attention": 0, "GroupNorm": 0, "SkipGroupNorm": 0, "NhwcConv": 17},
            "text_encoder": {
                "Attention": 5,
                "Gelu": 0,
                "LayerNormalization": 1,
                "QuickGelu": 5,
                "BiasGelu": 0,
                "SkipLayerNormalization": 10,
            },
        }

        export_onnx_dir = "tiny-random-sd"
        optimized_onnx_dir = "tiny-random-sd-optimized-fp32"
        # Disable GroupNorm due to limitation of current cuda kernel implementation.
        self.verify_pipeline_optimization(
            model_name,
            export_onnx_dir,
            optimized_onnx_dir,
            expected_op_counters,
            is_float16=False,
            atol=5e-3,
            disable_group_norm=True,
        )

        expected_op_counters["unet"].update({"Attention": 0, "MultiHeadAttention": 12})
        optimized_onnx_dir = "tiny-random-sd-optimized-fp16"
        self.verify_pipeline_optimization(
            model_name,
            export_onnx_dir,
            optimized_onnx_dir,
            expected_op_counters,
            is_float16=True,
            atol=5e-2,
            disable_group_norm=True,
        )

    @pytest.mark.slow
    def test_sdxl(self):
        """This tests optimization of SDXL pipeline"""
        model_name = TINY_MODELS["stable-diffusion-xl"]

        expected_op_counters = {
            "unet": {
                "Attention": 12,
                "MultiHeadAttention": 12,
                "LayerNormalization": 6,
                "SkipLayerNormalization": 30,
                "BiasSplitGelu": 0,
                "GroupNorm": 0,
                "SkipGroupNorm": 0,
                "NhwcConv": 35,
                "BiasAdd": 0,
            },
            "vae_encoder": {"Attention": 0, "GroupNorm": 0, "SkipGroupNorm": 0, "NhwcConv": 13},
            "vae_decoder": {"Attention": 0, "GroupNorm": 0, "SkipGroupNorm": 0, "NhwcConv": 17},
            "text_encoder": {
                "Attention": 5,
                "Gelu": 0,
                "LayerNormalization": 1,
                "QuickGelu": 0,
                "BiasGelu": 5,
                "SkipLayerNormalization": 10,
            },
            "text_encoder_2": {
                "Attention": 5,
                "Gelu": 0,
                "LayerNormalization": 1,
                "QuickGelu": 0,
                "BiasGelu": 5,
                "SkipLayerNormalization": 10,
            },
        }

        export_onnx_dir = "tiny-random-sdxl"
        optimized_onnx_dir = "tiny-random-sdxl-optimized-fp32"
        # Disable GroupNorm due to limitation of current cuda kernel implementation.
        self.verify_pipeline_optimization(
            model_name,
            export_onnx_dir,
            optimized_onnx_dir,
            expected_op_counters,
            is_float16=False,
            atol=5e-3,
            disable_group_norm=True,
        )

        expected_op_counters["unet"].update({"Attention": 0, "MultiHeadAttention": 24})
        optimized_onnx_dir = "tiny-random-sdxl-optimized-fp16"
        self.verify_pipeline_optimization(
            model_name,
            export_onnx_dir,
            optimized_onnx_dir,
            expected_op_counters,
            is_float16=True,
            atol=5e-2,
            disable_group_norm=True,
        )

    @pytest.mark.slow
    def test_sd3(self):
        """This tests optimization of stable diffusion 3 pipeline"""
        model_name = TINY_MODELS["stable-diffusion-3"]

        expected_op_counters = {
            "transformer": {
                "FastGelu": 3,
                "MultiHeadAttention": 2,
                "LayerNormalization": 8,
                "SimplifiedLayerNormalization": 0,
            },
            "vae_encoder": {"Attention": 0, "GroupNorm": 10, "SkipGroupNorm": 3, "NhwcConv": 17},
            "vae_decoder": {"Attention": 0, "GroupNorm": 14, "SkipGroupNorm": 7, "NhwcConv": 25},
            "text_encoder": {
                "Attention": 2,
                "Gelu": 0,
                "LayerNormalization": 1,
                "QuickGelu": 2,
                "SkipLayerNormalization": 4,
            },
            "text_encoder_2": {
                "Attention": 2,
                "Gelu": 0,
                "LayerNormalization": 1,
                "QuickGelu": 0,
                "SkipLayerNormalization": 4,
            },
            "text_encoder_3": {
                "Attention": 2,
                "MultiHeadAttention": 0,
                "Gelu": 0,
                "FastGelu": 2,
                "BiasGelu": 0,
                "GemmFastGelu": 0,
                "LayerNormalization": 0,
                "SimplifiedLayerNormalization": 2,
                "SkipLayerNormalization": 0,
                "SkipSimplifiedLayerNormalization": 3,
            },
        }

        export_onnx_dir = "tiny-random-stable-diffusion-3"
        optimized_onnx_dir = "tiny-random-stable-diffusion-3-optimized-fp32"
        self.verify_pipeline_optimization(
            model_name, export_onnx_dir, optimized_onnx_dir, expected_op_counters, is_float16=False, atol=5e-3
        )

        optimized_onnx_dir = "tiny-random-stable-diffusion-3-optimized-fp16"
        self.verify_pipeline_optimization(
            model_name, export_onnx_dir, optimized_onnx_dir, expected_op_counters, is_float16=True, atol=5e-2
        )

    @pytest.mark.slow
    def test_flux(self):
        """This tests optimization of flux pipeline"""
        model_name = TINY_MODELS["flux"]

        expected_op_counters = {
            "transformer": {
                "FastGelu": 8,
                "MultiHeadAttention": 6,
                "LayerNormalization": 13,
                "SimplifiedLayerNormalization": 16,
            },
            "vae_encoder": {"Attention": 0, "GroupNorm": 10, "SkipGroupNorm": 3, "NhwcConv": 17},
            "vae_decoder": {"Attention": 0, "GroupNorm": 14, "SkipGroupNorm": 7, "NhwcConv": 25},
            "text_encoder": {
                "Attention": 2,
                "Gelu": 0,
                "LayerNormalization": 1,
                "QuickGelu": 2,
                "SkipLayerNormalization": 4,
            },
            "text_encoder_2": {
                "Attention": 2,
                "MultiHeadAttention": 0,
                "Gelu": 0,
                "FastGelu": 2,
                "BiasGelu": 0,
                "GemmFastGelu": 0,
                "LayerNormalization": 0,
                "SimplifiedLayerNormalization": 2,
                "SkipLayerNormalization": 0,
                "SkipSimplifiedLayerNormalization": 3,
            },
        }

        export_onnx_dir = "tiny-random-flux"
        optimized_onnx_dir = "tiny-random-flux-optimized-fp32"
        self.verify_pipeline_optimization(
            model_name, export_onnx_dir, optimized_onnx_dir, expected_op_counters, is_float16=False, atol=1e-3
        )

        optimized_onnx_dir = "tiny-random-flux-optimized-fp16"
        self.verify_pipeline_optimization(
            model_name, export_onnx_dir, optimized_onnx_dir, expected_op_counters, is_float16=True, atol=5e-2
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import tempfile
import unittest
from pathlib import Path

import numpy as np
import onnx
from onnx import TensorProto, helper, numpy_helper
from op_test_utils import check_op_type_order

from onnxruntime.quantization.onnx_model import ONNXModel


def generate_input_initializer(tensor_shape, tensor_dtype, input_name):
    """
    Helper function to generate initializers for test inputs
    """
    tensor = np.random.normal(0, 0.3, tensor_shape).astype(tensor_dtype)
    init = numpy_helper.from_array(tensor, input_name)
    return init


def construct_model_for_topo_sort(model_path):
    #    (input)
    #       |
    #      GRU
    #      /  \
    #  Conv(1) \
    #     |     \
    #    Relu  Conv(2)
    #     |     |
    #     \     /
    #       Add
    #        |
    #       (output)
    initializers = []
    input = helper.make_tensor_value_info("input", TensorProto.FLOAT, [4, 8, 12])
    output = helper.make_tensor_value_info("output", TensorProto.FLOAT, [4, 2, 8, 8])

    # make GRU
    initializers.append(generate_input_initializer([2, 24, 12], np.float32, "W_GRU"))
    initializers.append(generate_input_initializer([2, 24, 8], np.float32, "R_GRU"))
    initializers.append(generate_input_initializer([2, 8, 8], np.float32, "H_GRU"))
    gru_node = helper.make_node(
        "GRU",
        ["input", "W_GRU", "R_GRU", "", "", "H_GRU"],
        ["GRU_O"],
        hidden_size=8,
        direction="bidirectional",
    )

    initializers.append(generate_input_initializer([2, 2, 1, 1], np.float32, "W1"))
    initializers.append(generate_input_initializer([2, 2, 1, 1], np.float32, "W2"))
    initializers.append(generate_input_initializer([2], np.float32, "B1"))
    initializers.append(generate_input_initializer([2], np.float32, "B2"))
    conv_node_1 = helper.make_node("Conv", ["GRU_O", "W1", "B1"], ["Conv1_O"], name="Conv1")
    conv_node_2 = helper.make_node("Conv", ["GRU_O", "W2", "B2"], ["Conv2_O"], name="Conv2")
    relu_node = helper.make_node("Relu", ["Conv1_O"], ["Relu_O"], name="Relu")
    add_node = helper.make_node("Add", ["Relu_O", "Conv2_O"], ["output"], name="Add")
    graph = helper.make_graph(
        [conv_node_1, relu_node, conv_node_2, gru_node, add_node],
        "onnx_model_test",
        [input],
        [output],
        initializer=initializers,
    )
    model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 13)])
    onnx.save(model, model_path)


def construct_model_for_topo_sort_constant(model_path):
    #    (input)    Constant
    #       \         /
    #        \       /
    #         \     /
    #          \   /
    #           Add
    #            |
    #         (output)

    initializers = []
    input = helper.make_tensor_value_info("input", TensorProto.FLOAT, [4, 8, 12])
    output = helper.make_tensor_value_info("output", TensorProto.FLOAT, [4, 8, 12])

    # make nodes
    constant_node = helper.make_node("Constant", [], ["const_output"], value_float=42.0)
    add_node = helper.make_node("Add", ["input", "const_output"], ["output"], name="Add")
    graph = helper.make_graph(
        [add_node, constant_node],
        "onnx_model_test",
        [input],
        [output],
        initializer=initializers,
    )
    model = helper.make_model(graph, opset_imports=[helper.make_opsetid("", 13)])
    onnx.save(model, model_path)


def construct_model_for_topo_sort_empty_input_output(model_path):
    #    (input1)    (input2)
    #       |           |
    #      Op1         Op1
    #       \         /
    #        \       /
    #         \     /
    #          \   /
    #           Op2
    #            |
    #           Op3
    #            |
    #         (output)

    input1 = helper.make_tensor_value_info("input1", TensorProto.FLOAT, [4, 8, 12])
    input2 = helper.make_tensor_value_info("input2", TensorProto.FLOAT, [4, 8, 12])
    output = helper.make_tensor_value_info("output", TensorProto.FLOAT, [4, 8, 12])

    # make nodes
    op1_node_1 = helper.make_node("Op1", ["input1"], ["", "", "Op1_1_output"], name="op1_1", domain="Test")
    op1_node_2 = helper.make_node("Op1", ["input2"], ["", "", "Op1_2_output"], name="op1_2", domain="Test")
    op2_node = helper.make_node("Op2", ["Op1_1_output", "Op1_2_output"], ["op2_output"], name="op2", domain="Test")
    op3_node = helper.make_node("Op3", ["", "op2_output"], ["output"], name="op3", domain="Test")
    graph = helper.make_graph(
        [op1_node_1, op1_node_2, op3_node, op2_node],
        "onnx_model_topo_test",
        [input1, input2],
        [output],
    )
    model = helper.make_model(graph, opset_imports=[helper.make_opsetid("Test", 1), helper.make_opsetid("", 13)])
    onnx.save(model, model_path)


class TestONNXModel(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._tmp_model_dir = tempfile.TemporaryDirectory(prefix="test_onnx_model.")

    @classmethod
    def tearDownClass(cls):
        cls._tmp_model_dir.cleanup()

    def test_topo_sort(self):
        test_model_path = str(Path(self._tmp_model_dir.name) / "onnx_model_topo_sort.onnx")
        construct_model_for_topo_sort(test_model_path)
        onnx_model = ONNXModel(onnx.load(test_model_path))
        check_op_type_order(self, onnx_model.model, ["Conv", "Relu", "Conv", "GRU", "Add"])
        onnx_model.topological_sort()
        check_op_type_order(self, onnx_model.model, ["GRU", "Conv", "Conv", "Relu", "Add"])

    def test_topo_sort_constant(self):
        test_model_path = str(Path(self._tmp_model_dir.name) / "onnx_model_topo_sort_constant.onnx")
        construct_model_for_topo_sort_constant(test_model_path)
        onnx_model = ONNXModel(onnx.load(test_model_path))
        check_op_type_order(self, onnx_model.model, ["Add", "Constant"])
        onnx_model.topological_sort()
        check_op_type_order(self, onnx_model.model, ["Constant", "Add"])

    def test_topo_sort_empty_input_output(self):
        test_model_path = str(Path(self._tmp_model_dir.name) / "onnx_model_topo_empty_input_output.onnx")
        construct_model_for_topo_sort_empty_input_output(test_model_path)
        onnx_model = ONNXModel(onnx.load(test_model_path))
        check_op_type_order(self, onnx_model.model, ["Op1", "Op1", "Op3", "Op2"])
        onnx_model.topological_sort()
        check_op_type_order(self, onnx_model.model, ["Op1", "Op1", "Op2", "Op3"])


if __name__ == "__main__":
    unittest.main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.
# --------------------------------------------------------------------------

# Running this script in Linux like the following
# CUBLAS_WORKSPACE_CONFIG=:4096:8 python multihead_attention_op_test_data_gen.py

import math

import numpy as np
import torch
from torch import Tensor, nn


# Modified from BertSelfAttention in https://github.com/huggingface/transformers/blob/main/src/transformers/models/bert/modeling_bert.py
class Attention(nn.Module):
    def __init__(
        self,
        num_attention_heads,
        hidden_dim,
        qk_head_size,
        v_head_size,
        is_decoder: bool,
    ):
        super().__init__()
        self.num_attention_heads = num_attention_heads
        self.qk_head_size = qk_head_size
        self.v_head_size = v_head_size
        self.qk_hidden_size = self.num_attention_heads * self.qk_head_size
        self.v_hidden_size = self.num_attention_heads * self.v_head_size

        self.query = nn.Linear(hidden_dim, self.qk_hidden_size)
        self.key = nn.Linear(hidden_dim, self.qk_hidden_size)
        self.value = nn.Linear(hidden_dim, self.v_hidden_size)
        self.is_decoder = is_decoder

        # Do not reshape output for pretty print.
        self.reshape_output = False
        self.verbose = False

    def transpose_for_scores(self, x: torch.Tensor, head_size) -> torch.Tensor:
        new_x_shape = x.size()[:-1] + (self.num_attention_heads, head_size)
        x = x.view(new_x_shape)
        return x.permute(0, 2, 1, 3)

    def get_extended_attention_mask(self, attention_mask: Tensor, dtype: torch.dtype) -> Tensor:
        assert attention_mask.dim() == 2 or attention_mask.dim() == 3
        extended_attention_mask = (
            attention_mask[:, None, :, :] if attention_mask.dim() == 3 else attention_mask[:, None, None, :]
        )
        extended_attention_mask = (1.0 - extended_attention_mask) * torch.finfo(dtype).min
        return extended_attention_mask

    def forward(
        self,
        hidden_states: torch.Tensor,
        attention_mask: torch.FloatTensor | None = None,
        encoder_hidden_states: torch.FloatTensor | None = None,
        encoder_attention_mask: torch.FloatTensor | None = None,
        past_key_value: tuple[tuple[torch.FloatTensor]] | None = None,
        output_attentions: bool | None = False,
    ) -> tuple[torch.Tensor]:
        mixed_query_layer = self.query(hidden_states)
        if self.verbose:
            print("q", mixed_query_layer)

        # If this is instantiated as a cross-attention module, the keys
        # and values come from an encoder; the attention mask needs to be
        # such that the encoder's padding tokens are not attended to.
        is_cross_attention = encoder_hidden_states is not None

        if is_cross_attention and past_key_value is not None:
            # reuse k,v, cross_attentions
            key_layer = past_key_value[0]
            value_layer = past_key_value[1]
            attention_mask = encoder_attention_mask
        elif is_cross_attention:
            k = self.key(encoder_hidden_states)
            if self.verbose:
                print("k", k)
            key_layer = self.transpose_for_scores(k, self.qk_head_size)
            if self.verbose:
                print("transposed key", key_layer)
            v = self.value(encoder_hidden_states)
            if self.verbose:
                print("v", v)
            value_layer = self.transpose_for_scores(v, self.v_head_size)
            if self.verbose:
                print("transposed value", value_layer)
            attention_mask = encoder_attention_mask
        elif past_key_value is not None:
            key_layer = self.transpose_for_scores(self.key(hidden_states), self.qk_head_size)
            value_layer = self.transpose_for_scores(self.value(hidden_states), self.v_head_size)
            key_layer = torch.cat([past_key_value[0], key_layer], dim=2)
            value_layer = torch.cat([past_key_value[1], value_layer], dim=2)
        else:
            key_layer = self.transpose_for_scores(self.key(hidden_states), self.qk_head_size)
            value_layer = self.transpose_for_scores(self.value(hidden_states), self.v_head_size)

        query_layer = self.transpose_for_scores(mixed_query_layer, self.qk_head_size)
        if self.verbose:
            print("transposed query", query_layer)

        if self.is_decoder:
            # if cross_attention save Tuple(torch.Tensor, torch.Tensor) of all cross attention key/value_states.
            # Further calls to cross_attention layer can then reuse all cross-attention
            # key/value_states (first "if" case)
            # if uni-directional self-attention (decoder) save Tuple(torch.Tensor, torch.Tensor) of
            # all previous decoder key/value_states. Further calls to uni-directional self-attention
            # can concat previous decoder key/value_states to current projected key/value_states (third "elif" case)
            # if encoder bi-directional self-attention `past_key_value` is always `None`
            past_key_value = (key_layer, value_layer)
            print("k cache", key_layer)
            print("v cache", value_layer)

        # Take the dot product between "query" and "key" to get the raw attention scores.
        attention_scores = torch.matmul(query_layer, key_layer.transpose(-1, -2))

        attention_scores = attention_scores / math.sqrt(self.qk_head_size)
        if self.verbose:
            print("QK", attention_scores)

        if attention_mask is not None:
            # Apply the attention mask is (precomputed for all layers in BertModel forward() function)
            attention_scores = attention_scores + self.get_extended_attention_mask(attention_mask, hidden_states.dtype)

        # Normalize the attention scores to probabilities.
        attention_probs = nn.functional.softmax(attention_scores, dim=-1)
        if self.verbose:
            print("softmax", attention_probs)

        context_layer = torch.matmul(attention_probs, value_layer)

        context_layer = context_layer.permute(0, 2, 1, 3).contiguous()

        if self.reshape_output:
            new_context_layer_shape = context_layer.size()[:-2] + (self.v_hidden_size,)
            context_layer = context_layer.view(new_context_layer_shape)

        print("output", context_layer)

        outputs = (context_layer, attention_probs) if output_attentions else (context_layer,)

        if self.is_decoder:
            outputs = (*outputs, past_key_value)

        return outputs


def run_cross_attention(
    hidden_dim,
    q_head_size,
    v_head_size,
    num_heads,
    batch_size,
    sequence_length,
    kv_sequence_length,
    key_padding_mask=None,
    has_bias=True,
    is_decoder=False,
):
    seed = 123
    torch.manual_seed(seed)
    np.random.seed(seed)
    torch.use_deterministic_algorithms(True)

    device = torch.device("cuda:0")
    mha = Attention(num_heads, hidden_dim, q_head_size, v_head_size, is_decoder=is_decoder).to(device).eval()
    if key_padding_mask is not None:
        key_padding_mask = key_padding_mask.to(device)
    torch.nn.init.uniform_(mha.query.weight, -0.5, 0.5)
    torch.nn.init.uniform_(mha.key.weight, -0.5, 0.5)
    torch.nn.init.uniform_(mha.value.weight, -0.5, 0.5)

    if has_bias:
        torch.nn.init.uniform_(mha.query.bias, -0.5, 0.5)
        torch.nn.init.uniform_(mha.key.bias, -0.5, 0.5)
        torch.nn.init.uniform_(mha.value.bias, -0.5, 0.5)
    else:
        torch.nn.init.zeros_(mha.query.bias)
        torch.nn.init.zeros_(mha.key.bias)
        torch.nn.init.zeros_(mha.value.bias)

    # Here we simulate input projection with MatMul but no bias:
    w_q = nn.Linear(hidden_dim, num_heads * q_head_size).to(device).eval()
    w_k = nn.Linear(hidden_dim, num_heads * q_head_size).to(device).eval()
    w_v = nn.Linear(hidden_dim, num_heads * v_head_size).to(device).eval()
    w_q.weight.copy_(mha.query.weight)
    w_k.weight.copy_(mha.key.weight)
    w_v.weight.copy_(mha.value.weight)
    torch.nn.init.zeros_(w_q.bias)
    torch.nn.init.zeros_(w_k.bias)
    torch.nn.init.zeros_(w_v.bias)

    torch.set_printoptions(profile="full", precision=8, linewidth=120, sci_mode=False)

    hidden_states = torch.empty(batch_size, sequence_length, hidden_dim, device="cuda")
    torch.nn.init.normal_(hidden_states)

    encoder_hidden_states = torch.empty(batch_size, kv_sequence_length, hidden_dim, device="cuda")
    torch.nn.init.normal_(encoder_hidden_states)

    input_q = w_q(hidden_states.clone())
    input_k = w_k(encoder_hidden_states.clone())
    input_v = w_v(encoder_hidden_states.clone())
    print("input_q", input_q)
    print("input_k", input_k)
    print("input_v", input_v)

    input_bias = torch.concat([mha.query.bias, mha.key.bias, mha.value.bias])
    print("input_bias", input_bias)
    if not has_bias:
        print("no bias!")

    # packed KV
    if q_head_size == v_head_size:
        packed_kv = torch.dstack(
            (
                input_k.reshape(batch_size * kv_sequence_length, num_heads, q_head_size),
                input_v.reshape(batch_size * kv_sequence_length, num_heads, v_head_size),
            )
        )
        packed_kv = packed_kv.reshape(batch_size, kv_sequence_length, num_heads, 2, q_head_size)
        print("packed_kv_5d", packed_kv)

    mha.forward(
        hidden_states,
        attention_mask=None,
        encoder_hidden_states=encoder_hidden_states,
        encoder_attention_mask=key_padding_mask,
        past_key_value=None,
        output_attentions=False,
    )

    if has_bias:
        print("zero out bias and run mha again")
        torch.nn.init.zeros_(mha.query.bias)
        torch.nn.init.zeros_(mha.key.bias)
        torch.nn.init.zeros_(mha.value.bias)
        mha.forward(
            hidden_states,
            attention_mask=None,
            encoder_hidden_states=encoder_hidden_states,
            encoder_attention_mask=key_padding_mask,
            past_key_value=None,
            output_attentions=False,
        )


def run_self_attention(
    hidden_dim,
    q_head_size,
    v_head_size,
    num_heads,
    batch_size,
    sequence_length,
    key_padding_mask=None,
    has_bias=True,
    is_decoder=False,
):
    seed = 123
    torch.manual_seed(seed)
    np.random.seed(seed)
    torch.use_deterministic_algorithms(True)

    device = torch.device("cuda:0")
    mha = Attention(num_heads, hidden_dim, q_head_size, v_head_size, is_decoder=is_decoder).to(device).eval()
    if key_padding_mask is not None:
        key_padding_mask = key_padding_mask.to(device)
    torch.nn.init.uniform_(mha.query.weight, -0.5, 0.5)
    torch.nn.init.uniform_(mha.key.weight, -0.5, 0.5)
    torch.nn.init.uniform_(mha.value.weight, -0.5, 0.5)

    if has_bias:
        torch.nn.init.uniform_(mha.query.bias, -0.5, 0.5)
        torch.nn.init.uniform_(mha.key.bias, -0.5, 0.5)
        torch.nn.init.uniform_(mha.value.bias, -0.5, 0.5)
    else:
        torch.nn.init.zeros_(mha.query.bias)
        torch.nn.init.zeros_(mha.key.bias)
        torch.nn.init.zeros_(mha.value.bias)

    # Here we simulate input projection with MatMul but no bias:
    w_q = nn.Linear(hidden_dim, num_heads * q_head_size).to(device).eval()
    w_k = nn.Linear(hidden_dim, num_heads * q_head_size).to(device).eval()
    w_v = nn.Linear(hidden_dim, num_heads * v_head_size).to(device).eval()
    w_q.weight.copy_(mha.query.weight)
    w_k.weight.copy_(mha.key.weight)
    w_v.weight.copy_(mha.value.weight)
    torch.nn.init.zeros_(w_q.bias)
    torch.nn.init.zeros_(w_k.bias)
    torch.nn.init.zeros_(w_v.bias)

    torch.set_printoptions(profile="full", precision=8, linewidth=120, sci_mode=False)

    hidden_states = torch.empty(batch_size, sequence_length, hidden_dim, device="cuda")
    torch.nn.init.normal_(hidden_states)

    input_q = w_q(hidden_states.clone())
    input_k = w_k(hidden_states.clone())
    input_v = w_v(hidden_states.clone())
    print("input_q", input_q)
    print("input_k", input_k)
    print("input_v", input_v)

    input_bias = torch.concat([mha.query.bias, mha.key.bias, mha.value.bias])
    print("input_bias", input_bias)
    if not has_bias:
        print("no bias!")

    # packed QKV
    if q_head_size == v_head_size:
        packed_qkv = torch.dstack(
            (
                input_q.reshape(batch_size * sequence_length, num_heads, q_head_size),
                input_k.reshape(batch_size * sequence_length, num_heads, q_head_size),
                input_v.reshape(batch_size * sequence_length, num_heads, v_head_size),
            )
        )
        packed_qkv = packed_qkv.reshape(batch_size, sequence_length, num_heads, 3, q_head_size)
        print("packed_qkv_5d", packed_qkv)

    mha.forward(
        hidden_states,
        attention_mask=None,
        encoder_hidden_states=None,
        encoder_attention_mask=key_padding_mask,
        past_key_value=None,
        output_attentions=False,
    )

    if has_bias:
        print("zero out bias and run mha again")
        torch.nn.init.zeros_(mha.query.bias)
        torch.nn.init.zeros_(mha.key.bias)
        torch.nn.init.zeros_(mha.value.bias)
        mha.forward(
            hidden_states,
            attention_mask=None,
            encoder_hidden_states=None,
            encoder_attention_mask=key_padding_mask,
            past_key_value=None,
            output_attentions=False,
        )


def run_cross_batch1_headsize_8():
    hidden_dim = 16
    q_head_size = 8
    v_head_size = 8
    num_heads = 2
    batch_size = 1
    sequence_length = 2
    kv_sequence_length = 3
    run_cross_attention(
        hidden_dim, q_head_size, v_head_size, num_heads, batch_size, sequence_length, kv_sequence_length
    )


def run_cross_batch2_headsize_40():
    hidden_dim = 80
    q_head_size = 40
    v_head_size = 40
    num_heads = 2
    batch_size = 2
    sequence_length = 3
    kv_sequence_length = 5
    run_cross_attention(
        hidden_dim, q_head_size, v_head_size, num_heads, batch_size, sequence_length, kv_sequence_length
    )


def run_cross_batch1_headsize_16():
    hidden_dim = 32
    q_head_size = 16
    v_head_size = 16
    num_heads = 2
    batch_size = 1
    sequence_length = 2
    kv_sequence_length = 3
    run_cross_attention(
        hidden_dim, q_head_size, v_head_size, num_heads, batch_size, sequence_length, kv_sequence_length
    )


def run_cross_batch2_headsize_16_8():
    hidden_dim = 32
    q_head_size = 16
    v_head_size = 8
    num_heads = 3
    batch_size = 2
    sequence_length = 1
    kv_sequence_length = 3
    run_cross_attention(
        hidden_dim, q_head_size, v_head_size, num_heads, batch_size, sequence_length, kv_sequence_length
    )


def run_cross_batch2_headsize_32_right_side_padding():
    hidden_dim = 64
    q_head_size = 32
    v_head_size = 32
    num_heads = 2
    batch_size = 2
    sequence_length = 2
    kv_sequence_length = 3
    key_padding_mask = torch.tensor([[1, 0, 0], [1, 1, 0]], dtype=torch.int32).cuda()

    run_cross_attention(
        hidden_dim,
        q_head_size,
        v_head_size,
        num_heads,
        batch_size,
        sequence_length,
        kv_sequence_length,
        key_padding_mask,
    )


def run_cross_batch1_headsize_32_left_side_padding():
    hidden_dim = 32
    q_head_size = 32
    v_head_size = 32
    num_heads = 1
    batch_size = 2
    sequence_length = 2
    kv_sequence_length = 3
    key_padding_mask = torch.tensor([[0, 1, 1], [0, 0, 1]], dtype=torch.int32).cuda()
    run_cross_attention(
        hidden_dim,
        q_head_size,
        v_head_size,
        num_heads,
        batch_size,
        sequence_length,
        kv_sequence_length,
        key_padding_mask,
    )


def run_cross_batch2_headsize_32_packed_kv():
    hidden_dim = 32
    q_head_size = 32
    v_head_size = 32
    num_heads = 1
    batch_size = 2
    sequence_length = 2
    kv_sequence_length = 3
    key_padding_mask = None
    has_bias = False
    run_cross_attention(
        hidden_dim,
        q_head_size,
        v_head_size,
        num_heads,
        batch_size,
        sequence_length,
        kv_sequence_length,
        key_padding_mask,
        has_bias,
    )


def run_self_batch2_headsize_32_packed_qkv():
    hidden_dim = 32
    q_head_size = 32
    v_head_size = 32
    num_heads = 1
    batch_size = 2
    sequence_length = 2
    key_padding_mask = None
    has_bias = False
    run_self_attention(
        hidden_dim, q_head_size, v_head_size, num_heads, batch_size, sequence_length, key_padding_mask, has_bias
    )


def run_cross_diff_seqlen_headsize_8():
    hidden_dim = 16
    q_head_size = 8
    v_head_size = 8
    num_heads = 2
    batch_size = 1
    sequence_length = 2
    kv_sequence_length = 4
    key_padding_mask = None
    has_bias = True
    run_cross_attention(
        hidden_dim,
        q_head_size,
        v_head_size,
        num_heads,
        batch_size,
        sequence_length,
        kv_sequence_length,
        key_padding_mask,
        has_bias,
        is_decoder=True,
    )


def run_self_past_present_headsize_8_nomask_no_attn_bias():
    hidden_dim = 16
    q_head_size = 8
    v_head_size = 8
    num_heads = 2
    batch_size = 2
    # In cpp side we use sequence_length = 1, we manually split the data of the first and second token.
    # Then we use first token related data as past and second token related data as true input.
    sequence_length = 2
    key_padding_mask = None
    has_bias = True
    run_self_attention(
        hidden_dim,
        q_head_size,
        v_head_size,
        num_heads,
        batch_size,
        sequence_length,
        key_padding_mask,
        has_bias,
        is_decoder=True,
    )


def create_test_data():
    """
    Create test data used in attention_op_test_helper.cc and multihead_attention_op_test.cc
    """
    print("CrossAttention_Batch1_HeadSize8")
    run_cross_batch1_headsize_8()

    print("CrossAttention_Batch2_HeadSize40")
    run_cross_batch2_headsize_40()

    print("CrossAttention_Batch1_HeadSize16")
    run_cross_batch1_headsize_16()

    print("CrossAttention_Batch2_HeadSize16_8")
    run_cross_batch2_headsize_16_8()

    print("CrossAttention_Batch2_HeadSize32_RightSidePadding")
    run_cross_batch2_headsize_32_right_side_padding()

    print("CrossAttention_Batch1_HeadSize32_LeftSidePadding")
    run_cross_batch1_headsize_32_left_side_padding()

    print("CrossAttention_Batch2_HeadSize32_PackedKV")
    run_cross_batch2_headsize_32_packed_kv()

    print("SelfAttention_Batch2_HeadSize32_PackedQKV")
    run_self_batch2_headsize_32_packed_qkv()

    print("SelfAttention_WithPastAndPresent_HeadSize8_NoMask_NoAttnBias")
    run_self_past_present_headsize_8_nomask_no_attn_bias()

    print("CrossAttention_DiffSequenceLengths_HeadSize8")
    run_cross_diff_seqlen_headsize_8()


with torch.no_grad():
    create_test_data()



import coremltools as ct
import numpy as np
from coremltools.converters.mil import Builder as mb
from coremltools.models import datatypes
from coremltools.models.neural_network import NeuralNetworkBuilder
from coremltools.models.utils import save_spec

input_dim = (1,)
output_dim = (1,)


def mlprogram():
    target = ct.target.iOS15

    @mb.program(input_specs=[mb.TensorSpec(shape=input_dim), mb.TensorSpec(shape=input_dim)], opset_version=target)
    def prog(x, y):
        return mb.real_div(x=x, y=y)

    # print(prog)

    # Convert to ML program
    m = ct.convert(prog, minimum_deployment_target=target)

    x = np.array([2], dtype=np.float32)
    y = np.array([2047], dtype=np.float32)

    # spec = m.get_spec()
    # print(spec)

    print(m.predict({"x": x, "y": y}))


# implement Div with coremltools approach of x * (1/y)
def nn():
    input_features = [("x", datatypes.Array(*input_dim)), ("y_inv", datatypes.Array(*input_dim))]
    output_features = [("final", datatypes.Array(*output_dim))]

    # Build a simple neural network with 1 inner product layer
    builder = NeuralNetworkBuilder(input_features, output_features)
    builder.add_elementwise(
        name="x_multiply_inverse_of_y",
        input_names=["x", "y_inv"],
        output_name="final",
        mode="MULTIPLY",
    )

    save_spec(builder.spec, "network.mlmodel")
    m = ct.models.MLModel("network.mlmodel")

    x = np.array([2], dtype=np.float32)
    y = np.array([1 / 2047], dtype=np.float32)
    print(m.predict({"x": x, "y_inv": y}))


def nn_scale():
    input_features = [
        ("x", datatypes.Array(*input_dim)),
        ("y_inv", datatypes.Array(*input_dim)),
        ("z", datatypes.Array(*input_dim)),
    ]
    output_features = [("final", datatypes.Array(*output_dim))]

    builder = NeuralNetworkBuilder(input_features, output_features)

    builder.add_elementwise(
        name="div_implemented_as_x_multiply_inverse_of_y",
        input_names=["x", "y_inv"],
        output_name="div_result",
        mode="MULTIPLY",
    )

    builder.add_elementwise(
        name="apply_scaling_factor",
        input_names=["div_result", "z"],
        output_name="final",
        mode="MULTIPLY",
    )

    from coremltools.models.utils import save_spec

    save_spec(builder.spec, "network.mlmodel")
    m = ct.models.MLModel("network.mlmodel")

    a = 2
    b = 2047
    # scaling factor to test working around coremltools inaccuracy.
    # weirdly even a scaling factor of 1 fixes the problem from https://github.com/microsoft/onnxruntime/issues/21170
    c = 1000

    x = np.array([a], dtype=np.float32)
    y = np.array([1 / b / c], dtype=np.float32)
    z = np.array([c], dtype=np.float32)
    print(m.predict({"x": x, "y_inv": y, "z": z}))


print("NN")
nn()

print("\nNN with scaling")
nn_scale()

print("\nML Program")
mlprogram()
